# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Settings class."""
from dataclasses import dataclass
from types import NoneType
from typing import Any

from qililab.typings import Parameter
from qililab.utils.castings import cast_enum_fields


@dataclass(kw_only=True)
class Settings:
    """Settings class."""

    def __post_init__(self):
        """Cast all enum attributes to its corresponding Enum class."""
        cast_enum_fields(obj=self)

    def set_parameter(self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None):
        """Cast the new value to its corresponding type and set the new attribute.

        Args:
            parameter (Parameter): Name of the parameter.
            value (float | bool | str): Value of the parameter.
            channel_id (int | None, optional): Channel id. Defaults to None.

        Raises:
            ValueError: If the parameter is a list and channel_id is None.
        """
        param: str = parameter.value
        attribute = getattr(self, param)

        if isinstance(attribute, list):
            self._set_parameter_to_attribute_list(value=value, attributes=attribute, channel_id=channel_id)
            return
        self._set_parameter_to_attribute(parameter_name=param, value=value, attribute=attribute)

    def _set_parameter_to_attribute(self, value: float | str | bool, parameter_name: str, attribute: Any):
        """Set the parameter value to its corresponding attribute

        Args:
            value (float | str | bool): _description_
            parameter_name (str): _description_
            attribute (Any): _description_
        """
        attr_type = type(attribute)
        if attr_type == int:  # FIXME: Depending on how we define de value, python thinks it is an int
            attr_type = float
        if attr_type != NoneType:
            value = attr_type(value)
        setattr(self, parameter_name, value)

    def _set_parameter_to_attribute_list(
        self,
        value: float | str | bool,
        attributes: list[float | str | bool],
        channel_id: int | None,
    ):
        """Set the parameter value to its corresponding attribute list element

        Args:
            value (float | str | bool): _description_
            attribute (list[float  |  str  |  bool]): _description_
            channel_id (int | None): _description_
        """
        if channel_id is None:
            raise ValueError("No list index specified when updating a list of parameters.")
        if len(attributes) <= channel_id:
            raise ValueError(
                "Index out of bounds: Trying to update a list of parameters with length "
                + f"{len(attributes)} at position {channel_id}"
            )
        attributes[channel_id] = value

    def get_parameter(self, parameter: Parameter, channel_id: int | None = None):
        """Get parameter from settings.

        Args:
            parameter (Parameter): Name of the parameter.
            channel_id (int | None, optional): Channel id. Defaults to None.

        Raises:
            ValueError: If the parameter is a list and channel_id is None.

        Returns:
            int | float | bool | str: Value of the parameter.
        """
        param: str = parameter.value
        attribute = getattr(self, param)

        if isinstance(attribute, list):
            if channel_id is None:
                raise ValueError(f"channel_id must be specified to get parameter {param}.")
            return attribute[channel_id]
        return attribute



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is the top level module from which all basic functions and classes of
Qililab can be directly imported.
"""

# isort: skip_file
import contextlib

from .about import about
from .config import __version__, logger
from .data_management import build_platform, load_results, save_platform, save_results
from .execute_circuit import execute
from .experiment import Experiment
from .qprogram import Domain, QbloxCompiler, QProgram, QuantumMachinesCompiler
from .result import Results, stream_results
from .typings import ExperimentOptions, ExperimentSettings, Parameter
from .utils import Loop
from .utils.load_data import load
from .waveforms import *

# moving circuit_transpiler module imports here because it has instruments module dependencies so circular imports can be avoided
from .circuit_transpiler import Drag, Wait

with contextlib.suppress(NameError, ImportError):
    # Since Ipython magic methods can only be imported from inside a Jupyter Notebook,
    # here we first check that `get_ipython` exists (which means we are inside a Jupyter Notebook)
    get_ipython()  # type: ignore  # noqa: F405 # pylint: disable=undefined-variable
    from .slurm import submit_job  # pragma: no cover



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Driver for the Drive Bus class."""
from qililab.drivers.interfaces.attenuator import Attenuator
from qililab.drivers.interfaces.awg import AWG
from qililab.drivers.interfaces.local_oscillator import LocalOscillator
from qililab.platform.components.bus_driver import BusDriver
from qililab.platform.components.bus_factory import BusFactory


@BusFactory.register
class DriveBus(BusDriver):
    """Qililab's driver for Drive Bus.

    Args:
        alias: Bus alias.
        port (int): Port to target.
        awg (AWG): Sequencer.
        local_oscillator (LocalOscillator | None): Local oscillator.
        attenuator (Attenuator | None): Attenuator.
        distortions (list): Distortions to apply in this Bus.

    Returns:
        BusDriver: BusDriver instance of type drive bus.
    """

    def __init__(
        self,
        alias: str,
        port: int,
        awg: AWG,
        local_oscillator: LocalOscillator | None,
        attenuator: Attenuator | None,
        distortions: list,
    ):
        super().__init__(alias=alias, port=port, awg=awg, distortions=distortions)
        if local_oscillator:
            self.instruments["local_oscillator"] = local_oscillator
        if attenuator:
            self.instruments["attenuator"] = attenuator



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bus Class Interface."""
from abc import ABC
from copy import deepcopy
from typing import Any

from qililab.drivers.interfaces import AWG, BaseInstrument
from qililab.drivers.interfaces.instrument_interface_factory import InstrumentInterfaceFactory
from qililab.pulse import PulseBusSchedule, PulseDistortion


class BusDriver(ABC):
    """Derived: :class:`DriveBus`, :class:`FluxBus` and :class:`ReadoutBus`

    Bus abstract base class.

    Args:
        alias (str): Bus alias.
        port (int): Port to target.
        awg (AWG): Sequencer.
        distortions (list): Distortions to apply in this Bus.
    """

    def __init__(self, alias: str, port: int, awg: AWG | None, distortions: list):
        self.alias = alias
        self.port = port
        self._awg = awg
        self.instruments: dict[str, BaseInstrument | None] = {"awg": self._awg}
        self.delay: int = 0
        self.distortions: list[PulseDistortion] = [
            PulseDistortion.from_dict(distortion) if isinstance(distortion, dict) else distortion
            for distortion in distortions
        ]

    def execute(
        self,
        pulse_bus_schedule: PulseBusSchedule,
        nshots: int,
        repetition_duration: int,
        num_bins: int,
    ) -> None:
        """Execute a pulse bus schedule through the AWG Instrument.

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse Bus Schedule to generate QASM program.
            nshots (int): number of shots
            repetition_duration (int): repetition duration.
            num_bins (int): number of bins
        """
        if self._awg:
            self._awg.execute(
                pulse_bus_schedule=pulse_bus_schedule,
                nshots=nshots,
                repetition_duration=repetition_duration,
                num_bins=num_bins,
            )

    def set(self, param_name: str, value: Any) -> None:
        """Set parameter on the bus' instruments.

        Args:
            param (str): Parameter's name.
            value (Any): Parameter's value

        Raises:
            AttributeError: if more than one instrument has the same parameter name.
            AttributeError: if no instrument is found for the parameter name.
        """
        if param_name == "delay":
            self.delay = value
        elif param_name == "distortions":
            raise NotImplementedError("Setting distortion parameters of a bus is not yet implemented..")
        else:
            candidates: list[BaseInstrument | None] = [
                instrument for instrument in self.instruments.values() if instrument and param_name in instrument.params
            ]
            if len(candidates) == 1 and isinstance(candidates[0], BaseInstrument):
                candidates[0].set(param_name, value)
            elif len(candidates) == 2 and candidates[0] == candidates[1] and isinstance(candidates[0], BaseInstrument):
                candidates[0].set(param_name, value)
            elif len(candidates) > 1:
                raise AttributeError(f"Bus {self.alias} contains multiple instruments with the parameter {param_name}.")
            else:
                raise AttributeError(
                    f"Bus {self.alias} doesn't contain any instrument with the parameter {param_name}."
                )

    def get(self, param_name: str) -> Any:
        """Return value associated to a parameter on the bus' instrument.

        Args:
            param (str): Parameter's name.

        Returns:
            value (Any): Parameter's value

        Raises:
            AttributeError: if more than one instrument has the same parameter name.
            AttributeError: if no instrument is found for the parameter name.
        """
        if param_name == "delay":
            return self.delay
        if param_name == "distortions":
            raise NotImplementedError("Getting distortion parameters of a bus is not yet implemented.")
        candidates: list[BaseInstrument | None] = [
            instrument for instrument in self.instruments.values() if instrument and param_name in instrument.params
        ]
        if len(candidates) == 1 and isinstance(candidates[0], BaseInstrument):
            return candidates[0].get(param_name)
        if len(candidates) == 2 and candidates[0] == candidates[1] and isinstance(candidates[0], BaseInstrument):
            return candidates[0].get(param_name)
        if len(candidates) > 1:
            raise AttributeError(f"Bus {self.alias} contains multiple instruments with the parameter {param_name}.")
        raise AttributeError(f"Bus {self.alias} doesn't contain any instrument with the parameter {param_name}.")

    @classmethod
    def from_dict(cls, dictionary: dict, instruments: list[BaseInstrument]) -> "BusDriver":
        """Loads the corresponding BusDriver class and sets the instrument parameters for the corresponding instruments.

        The input dictionary should conform to the following structure: [https://imgur.com/a/U4Oyapo]

        .. code-block:: yaml

            alias: bus_example
            type: DriveBus
            AWG:
                alias: q0_readout
                parameters:
                    gain: 0.9
                    ...
            LO:
                alias: lo_readout
                parameters:
                    frequency: 1e9
                    ...
            Attenuator:
                alias: attenuator_0
                parameters:
                    attenuation: 20
                    ...
            port: 100
            distortions: []

        To correctly set the instrument parameters, the instrument alias (`q0_readout`, ...) and its corresponding
        interface (`AWG`, `LO`, ...) need to match with one of the passed instruments.

        If an instrument serves multiple interfaces, its parameters should be set through only one interface to avoid
        redundancy or to avoid setting them multiple times with different values. However, the (same) alias should be
        specified in all corresponding interfaces.

        The interfaces specified in the dictionary must adhere to the format and sequence outlined in the keys of
        `instrument_interfaces_caps_translate()`. Valid interface types are: AWG, Digitiser, LocalOscillator,
        Attenuator, VoltageSource, and CurrentSource.

        Args:
            dictionary (dict): A dictionary representing the BusDriver object and its instrument parameters.
            instruments (list[BaseInstrument]): A list of pre-instantiated instrument objects.

        Returns:
            BusDriver: The initialized BusDriver class.
        """
        from .bus_factory import BusFactory  # pylint: disable=import-outside-toplevel, cyclic-import

        local_dictionary = deepcopy(dictionary)
        local_dictionary.pop("type", None)

        local_dictionary = cls.__convert_instruments_strings_to_classes_and_set_params(
            dictionary=local_dictionary, instruments=instruments
        )

        return BusFactory.get(name=dictionary["type"])(**local_dictionary)  # type: ignore[return-value]

    def to_dict(self) -> dict:
        """Generates a dictionary representation of the Bus and its corresponding instrument parameters.

        The structure of the output dictionary is the following: [https://imgur.com/a/U4Oyapo]

        .. code-block:: yaml

            alias: bus_example
            type: DriveBus
            AWG:
                alias: q0_readout
                parameters:
                    gain: 0.9
                    ...
            LO:
                alias: lo_readout
                parameters:
                    frequency: 1e9
                    ...
            Attenuator:
                alias: attenuator_0
                parameters:
                    attenuation: 20
                    ...
            port: 100
            distortions: []

        The interfaces in the runcard are formatted and ordered according to the keys of `instrument_interfaces_caps_translate()`.
        Acceptable interface types are: AWG, Digitiser, LocalOscillator, Attenuator, VoltageSource, and CurrentSource.

        If an instrument serves multiple interfaces, its parameters will only be included under the first interface
        that appears in the list. However, the alias for that instrument will be repeated across all its interfaces,
        making it easier to associate them.

        The sequence defined in `instrument_interfaces_caps_translate()` is significant because it dictates the order
        in which the parameters are written. Since each instrument's parameters are only saved once, specifically under
        the first interface that appears, the sequence will determine which interface's dictionary entry will include both
        the parameters and the alias, and which will include only the alias.

        Returns:
            dict: A dictionary representing the Bus and its associated instrument parameters.
        """
        bus_dictionary_start: dict = {
            "alias": self.alias,
            "type": self.__class__.__name__,
        }

        instruments_dictionary = self.__convert_instruments_classes_to_strings_and_get_params(
            instruments=[instrument for instrument in self.instruments.values() if instrument is not None]
        )

        bus_dictionary_end = {
            "port": self.port,
            "distortions": [distortion.to_dict() for distortion in self.distortions],
        }

        return bus_dictionary_start | instruments_dictionary | bus_dictionary_end

    @classmethod
    def __convert_instruments_strings_to_classes_and_set_params(
        cls, dictionary: dict, instruments: list[BaseInstrument]
    ) -> dict:
        """Function called in the `from_dict()` method to construct an instance of a bus class from a dictionary.

        Passes the strings of the instruments associated to the bus, into their corresponding (already instantiated)
        classes, through their "alias" and interface. While it also sets their corresponding given parameters.

        To do so, we translate the caps of the instrument interfaces with `instrument_interfaces_caps_translate()`
        (from "AWG" to "awg", from "LocalOscillator" to "local_oscillator", etc...).

        And finally returns a dictionary with only the instruments classes.

        The dictionary reading follows the following structure: [https://imgur.com/a/U4Oyapo], check th `from_dict()`
        documentation for more information regarding this.

        Args:
            dictionary (dict): A dictionary representing the BusDriver object and its instrument parameters.
            instruments (list[BaseInstrument]): A list of pre-instantiated instrument objects.

        Returns:
            dict: Bus dictionary with the instruments as classes to be inserted in the bus dictionary.
        """
        instruments_dictionary: dict[str, BaseInstrument] = {}
        used_keys: list[str] = []

        for key, instrument_dict in dictionary.items():  # pylint: disable=too-many-nested-blocks
            if key in cls.__instrument_interfaces_caps_translate():
                for instrument in instruments:
                    # If the alias and the interface of the dictionary coincide with one of the given instruments:
                    if (
                        issubclass(instrument.__class__, InstrumentInterfaceFactory.get(key))
                        and instrument.alias == instrument_dict["alias"]
                    ):
                        # Set parameters of the initialized instrument
                        if "parameters" in instrument_dict:
                            for parameter, value in instrument_dict["parameters"].items():
                                instrument.set(parameter, value)

                        # Save the instrument classes in the corresponding key, after translating the caps of the interfaces.
                        instruments_dictionary[cls.__instrument_interfaces_caps_translate()[key]] = instrument
                        break

                # Remember used keys, to remove the instrument strings from the original dictionary
                used_keys.append(key)

        for key in used_keys:
            dictionary.pop(key)

        # Add new initialize instrument classes to dictionary
        return dictionary | instruments_dictionary

    @classmethod
    def __convert_instruments_classes_to_strings_and_get_params(
        cls, instruments: list[BaseInstrument]
    ) -> dict[str, dict]:
        """Function called in the `to_dict()` method, to construct a dictionary representing the actual bus instance.

        Passes the instruments classes associated to the bus, into their corresponding strings. While it
        also gets all their corresponding parameters to be printed together in a dictionary.

        To do so, we use the caps format of the instrument interfaces from `instrument_interfaces_caps_translate()`.
        And the order is important here, because it marks the order of writing, and since we only save the parameters
        of each instrument once, concretely in the first interface to appear, this means that the order of
        `__instrument_interfaces_caps_translate()` will indicate in which interface dict you save the
        parameters & alias and in which you only save the alias.

        And finally returns a dictionary with the instrument string as key (str), and the alias and parameters as values (dict).

        The dictionary construction follows the following structure: [https://imgur.com/a/U4Oyapo], check th `to_dict()`
        documentation for more information regarding this.

        Args:
            instruments (list[BaseInstrument]): Instruments corresponding the the given Bus.

        Returns:
            dict[str, dict]: The instruments dictionary to be inserted in the bus dictionary. Keys are the instruments
                strings, and values are an inner dictionary containing the alias and the parameters dictionary.
        """
        instruments_dict: dict[str, dict] = {}
        saved_instruments: set[str] = set()

        # The order of keys marks the writting order and in which interface dict you save the parameters & alias and in which only the alias.
        for key in cls.__instrument_interfaces_caps_translate():
            for instrument in instruments:
                if issubclass(instrument.__class__, InstrumentInterfaceFactory.get(key)):
                    # Add alias of the instrument to the dictionary
                    instruments_dict[key] = {"alias": instrument.alias}

                    # This ensures that instruments with multiple interfaces, don't write the same parameters two times
                    if instrument.alias not in saved_instruments and instrument.params:
                        # Add parameters of the instrument to the dictionary
                        instruments_dict[key]["parameters"] = {
                            parameter: instrument.get(parameter)
                            for parameter in instrument.params.keys()
                            if parameter
                            not in (
                                "IDN",
                                "sequence",
                            )  # skip IDN and sequence parameters. Which will go into other parts of the runcard.
                        }

                    # Save already saved instruments, to not write same parameters twice (in different interfaces)
                    saved_instruments.add(instrument.alias)
                    break

        return instruments_dict

    @staticmethod
    def __instrument_interfaces_caps_translate() -> dict:
        """Dictionary to translate the instruments[str] to the caps showing in the Runcard (for aesthetics).

        The order of the interfaces here determine the order for the printing too, and since we only save the parameters
        of a same instrument once, concretely in the first interface to appear, this means that the order also indicates
        in which interface dict you save the parameters & alias and in which you only save the alias. .
        """
        return {
            "AWG": "awg",
            "Digitiser": "digitiser",
            "LocalOscillator": "local_oscillator",
            "Attenuator": "attenuator",
            "VoltageSource": "source",
            "CurrentSource": "source",
        }

    def __eq__(self, other: object) -> bool:
        """compare two Bus objects"""
        return str(self) == str(other) if isinstance(other, BusDriver) else False

    def __str__(self):
        """String representation of a Bus."""
        return (
            f"{self.alias} ({self.__class__.__name__}): "
            + "".join(f"--|{instrument.alias}|" for instrument in self.instruments.values())
            + f"--> port {self.port}"
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Platform class."""
import ast
import io
import re
from copy import deepcopy
from dataclasses import asdict
from queue import Queue

from qibo.gates import M
from qibo.models import Circuit
from qiboconnection.api import API
from qm import generate_qua_script
from qpysequence import Sequence as QpySequence
from ruamel.yaml import YAML

from qililab.chip import Chip
from qililab.circuit_transpiler import CircuitTranspiler
from qililab.config import logger
from qililab.constants import GATE_ALIAS_REGEX, RUNCARD
from qililab.instrument_controllers import InstrumentController, InstrumentControllers
from qililab.instrument_controllers.utils import InstrumentControllerFactory
from qililab.instruments.instrument import Instrument
from qililab.instruments.instruments import Instruments
from qililab.instruments.qblox import QbloxModule
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.instruments.utils import InstrumentFactory
from qililab.pulse import PulseSchedule
from qililab.pulse import QbloxCompiler as PulseQbloxCompiler
from qililab.qprogram import QbloxCompiler, QProgram, QuantumMachinesCompiler
from qililab.result import Result
from qililab.result.qblox_results.qblox_result import QbloxResult
from qililab.result.qprogram.qprogram_results import QProgramResults
from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult
from qililab.settings import Runcard
from qililab.system_control import ReadoutSystemControl
from qililab.typings.enums import InstrumentName, Line, Parameter
from qililab.utils import hash_qpy_sequence

from .components import Bus, Buses


class Platform:  # pylint: disable = too-many-public-methods, too-many-instance-attributes
    """Platform object representing the laboratory setup used to control quantum devices.

    The platform is responsible for managing the initializations, connections, setups, and executions of the laboratory, which mainly consists of:

    - :class:`.Chip`

    - Buses

    - Instruments

    .. note::

        This class should be instantiated with the :meth:`ql.build_platform()` function, either by passing a :ref:`runcard <runcards>` (serialized platform dictionary)
        or a path to the location of the YAML file containing it.

        More information about the runcard structure, in the :ref:`Runcards <runcards>` section of the documentation.

    After initializing a :class:`Platform`, the typical first three steps (which are usually only required at the start) are:

    >>> platform.connect() # Connects to all the instruments.
    >>> platform.initial_setup()  # Sets the parameters defined in the runcard.
    >>> platform.turn_on_instruments()  # Turns on the signal outputs.

    And then, for each experiment you want to run, you would typically repeat:

    >>> platform.set_parameter(...) # Sets any parameter of the Platform.
    >>> result = platform.execute(...) # Executes the platform.

    Args:
        runcard (Runcard): Dataclass containing the serialized platform (chip, instruments, buses...), created during :meth:`ql.build_platform()` with the given runcard dictionary.
        connection (API | None = None): `Qiboconnection's <https://pypi.org/project/qiboconnection>`_ API class used to block access to other users when connected to the platform.

    Examples:

        .. note::

            The following examples contain made up results. These will soon be updated with real results.

        .. note::

            All the following examples are explained in detail in the :ref:`Platform <platform>` section of the documentation. However, here are a few thing to keep in mind:

            - To connect, your computer must be in the same network of the instruments specified in the runcard, with their IP's addresses. Connection is necessary for the subsequent steps.

            - You might want to skip the ``platform.initial_setup()`` and the ``platform.turn_on_instruments()`` steps if you think nothing has been modified, but we recommend doing them every time.

            - ``platform.turn_on_instruments()`` is used to turn on the signal output of all the sources defined in the runcard (RF, Voltage and Current sources).

            - You can print ``platform.chip`` and ``platform.buses`` at any time to check the platform's structure.

        **1. Executing a circuit with Platform:**


        To execute a circuit you first need to define your circuit, for example, one with a pi pulse and a measurement gate in qubit ``q`` (``int``).
        Then you also need to build, connect, set up, and execute the platform, which together look like:

        .. code-block:: python

            import qililab as ql

            from qibo.models import Circuit
            from qibo import gates

            # Defining the Rabi circuit:
            circuit = Circuit(q+1)
            circuit.add(gates.X(q))
            circuit.add(gates.M(q))

            # Building the platform:
            platform = ql.build_platform(runcard="runcards/galadriel.yml")

            # Connecting and setting up the platform:
            platform.connect()
            platform.initial_setup()
            platform.turn_on_instruments()

            # Executing the platform:
            result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)

        The results would look something like this:

        >>> result.array
        array([[6.],
                [6.]])

        .. note::

            The obtained values correspond to the integral of the I/Q signals received by the digitizer.
            And they have shape `(#sequencers, 2, #bins)`, in this case you only have 1 sequencer and 1 bin.

        You could also get the results in a more standard format, as already classified ``counts`` or ``probabilities`` dictionaries, with:

        >>> result.counts
        {'0': 501, '1': 499}

        >>> result.probabilities
        {'0': .501, '1': .499}

        .. note::

            You can find more information about the results, in the :class:`.Results` class documentation.

        |

        **2. Running a Rabi sweep with Platform:**

        To perform a Rabi sweep, you need the previous circuit, and again, you also need to build, connect and setup the platform.
        But this time, instead than executing the circuit once, you will loop changing the amplitude parameter of the AWG (generator of the pi pulse):

        .. code-block:: python

            # Looping over the AWG amplitude to execute the Rabi sweep:
            results = []
            amp_values = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.9, 1.0]

            for amp in amp_values:
                platform.set_parameter(alias="drive_q", parameter=ql.Parameter.AMPLITUDE, value=amp)
                result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        Now you can use ``np.hstack`` to stack the results horizontally. By doing this, you would obtain an
        array with shape `(2, N)`, where N is the number of elements inside the loop:

        >>> import numpy as np
        >>> np.hstack(results)
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])

        You can see how the integrated I/Q values oscillate, indicating that qubit ``q`` oscillates between the ground and
        excited states!

        |

        **3. A faster Rabi sweep, translating the circuit to pulses:**

        Since you are looping over variables that are independent of the circuit (in this case, the amplitude of the AWG),
        you can speed up the experiment by translating the circuit into pulses beforehand, only once, and then, executing the obtained
        pulses inside the loop.

        Which is the same as before, but passing the ``pulse_schedule`` instead than the ``circuit``, to the ``execute()`` method:

        .. code-block:: python

            from qililab.pulse.circuit_to_pulses import CircuitToPulses

            # Translating the circuit to pulses:
            pulse_schedule = CircuitToPulses(platform=platform).translate(circuits=[circuit])

            # Looping over the AWG amplitude to execute the Rabi sweep:
            results = []
            amp_values = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.9, 1.0]

            for amp in amp_values:
                platform.set_parameter(alias="drive_q", parameter=ql.Parameter.AMPLITUDE, value=amp)
                result = platform.execute(program=pulse_schedule, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        If you now stack and print the results, you will obtain similar results, but much faster!

        >>> np.hstack(results)
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])
        TODO: !!! Change this results for the actual ones !!!

        |

        **4. Ramsey sequence, looping over a parameter inside the circuit:**

        To run a Ramsey sequence you also need to build, connect and set up the platform as before. However, the circuit will be different from the previous one,
        and also, this time, you need to loop over a parameter of the circuit itself, specifically over the time of the ``Wait`` gate.

        To do this, since the parameter is inside the Qibo circuit, you will need to use Qibo own ``circuit.set_parameters()`` method, specifying the
        parameters you want to set, in the same order they appear in the circuit construction:

        .. code-block:: python

            import qililab as ql

            from qibo.models import Circuit
            from qibo import gates

            # Building the platform:
            platform = ql.build_platform(runcard="runcards/galadriel.yml")

            # Connecting and setting up the platform:
            platform.connect()
            platform.initial_setup()
            platform.turn_on_instruments()

            # Defining the Ramsey circuit:
            circuit = Circuit(q + 1)
            circuit.add(gates.RX(q, theta=np.pi/2))
            circuit.add(ql.Wait(q, t=0))
            circuit.add(gates.RX(q, theta=np.pi/2))
            circuit.add(gates.M(q))

            # Looping over the wait time t to execute the Ramsey:
            results = []
            wait_times = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

            for wait in wait_times:
                circuit.set_parameters([np.pi/2, wait, np.pi/2])
                result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        which for each execution, would set ``np.pi/2`` to the ``theta`` parameters of the ``RX`` gates, and the looped ``wait`` time  to the ``t`` parameter of the
        ``Wait`` gate.

        If you print the results, you'll see how you obtain the sinusoidal expected behaviour!

        >>> results = np.hstack(results)
        >>> results
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])
        TODO: !!! Change this results for the actual sinusoidal ones (change wait_times of execution if needed) !!!
    """

    def __init__(self, runcard: Runcard, connection: API | None = None):
        self.name = runcard.name
        """Name of the platform (``str``) """

        self.device_id = runcard.device_id
        """Device ID of the platform (``int``). This attribute is required for ``qiboconnection`` to save the results remotely."""

        self.gates_settings = runcard.gates_settings
        """Gate settings and definitions (``dataclass``). These setting contain how to decompose gates into pulses."""

        self.instruments = Instruments(elements=self._load_instruments(instruments_dict=runcard.instruments))
        """All the instruments of the platform and their necessary settings (``dataclass``). Each individual instrument is contained in a list within the dataclass."""

        self.instrument_controllers = InstrumentControllers(
            elements=self._load_instrument_controllers(instrument_controllers_dict=runcard.instrument_controllers)
        )
        """All the instrument controllers of the platform and their necessary settings (``dataclass``). Each individual instrument controller is contained in a list within the dataclass."""

        self.chip = Chip(**asdict(runcard.chip))
        """Chip and nodes settings of the platform (:class:`.Chip` dataclass). Each individual node is contained in a list within the :class:`.Chip` class."""

        self.buses = Buses(
            elements=[
                Bus(settings=asdict(bus), platform_instruments=self.instruments, chip=self.chip)
                for bus in runcard.buses
            ]
        )
        """All the buses of the platform and their necessary settings (``dataclass``). Each individual bus is contained in a list within the dataclass."""

        self.connection = connection
        """API ``qiboconnection`` of the platform (``API | None``). It is the same as the passed argument. Defaults to None."""

        self._connected_to_instruments: bool = False
        """Boolean indicating the connection status to the instruments. Defaults to False (not connected)."""

        if any(isinstance(instrument, QbloxModule) for instrument in self.instruments.elements):
            self.compiler = PulseQbloxCompiler(platform=self)  # TODO: integrate with qprogram compiler
            """Compiler to translate given programs to instructions for a given awg vendor."""

        self._qpy_sequence_cache: dict[str, str] = {}
        """Dictionary for caching qpysequences."""

    def connect(self, manual_override=False):
        """Connects to all the instruments and blocks the connection for other users.

        You must be connected in order to set up and turn on instruments, or in order to execute the platform.

        To connect, your computer must be in the same network of the instruments specified in the :ref:`runcard <runcards>` (with their corresponding `device_id` and IP's addresses).

        Such connection is handled via `qiboconnection's <https://pypi.org/project/qiboconnection>`_ `API` in the ``platform.connection`` attribute.

        Args:
            manual_override (bool, optional): If ``True``, avoid checking if the device is blocked (surpasses any blocked connection). This will stop any
                current execution. Defaults to False.
        """
        if self._connected_to_instruments:
            logger.info("Already connected to the instruments")
            return

        if self.connection is not None and not manual_override:
            self.connection.block_device_id(device_id=self.device_id)

        self.instrument_controllers.connect()
        self._connected_to_instruments = True
        logger.info("Connected to the instruments")

    def initial_setup(self):
        """Sets the values of the cache of the :class:`.Platform` object to the connected instruments.

        If called after a ``ql.build_platform()``, where the :class:`.Platform` object is built with the provided runcard,
        this function sets the values of the :ref:`runcard <runcards>` into the connected instruments.

        It is recommended to use this function after a ``ql.build_platform()`` + ``platform.connect()`` to ensure that no parameter
        differs from the current runcard settings.

        If a `platform.set_parameter()` is called between platform building and initial setup, the value set in the instruments
        will be the new "set" value, as the cache values of the :class:`.Platform` object are modified.
        """
        if not self._connected_to_instruments:
            raise AttributeError("Can not do initial_setup without being connected to the instruments.")
        self.instrument_controllers.initial_setup()
        logger.info("Initial setup applied to the instruments")

    def turn_on_instruments(self):
        """Turns on the signal output for the generator instruments (RF, voltage sources and current sources).

        This does not actually turn on the laboratory instruments, it only opens the signal output generation of the sources.

        We recommend you to do this always after a connection and a setup, to ensure that everything is ready for an execution.
        """
        self.instrument_controllers.turn_on_instruments()
        logger.info("Instruments turned on")

    def turn_off_instruments(self):
        """Turns off the signal output for the generator instruments (local oscillators, voltage sources and current sources).

        This does not actually turn the laboratory instruments off, it only closes their signal output generation.
        """
        self.instrument_controllers.turn_off_instruments()
        logger.info("Instruments turned off")

    def disconnect(self):
        """Closes the connection to all the instruments."""
        if self.connection is not None:
            self.connection.release_device(device_id=self.device_id)
        if not self._connected_to_instruments:
            logger.info("Already disconnected from the instruments")
            return
        self.instrument_controllers.disconnect()
        self._connected_to_instruments = False
        logger.info("Disconnected from instruments")

    def get_element(self, alias: str):
        """Gets the platform element and to which bus it is connected, using its alias.

        Args:
            alias (str): Element alias to identify it.

        Returns:
            tuple[object, list | None]: Element class together with the index of the bus where the element is located.
        """
        if alias is not None:
            if alias == "platform":
                return self.gates_settings
            regex_match = re.search(GATE_ALIAS_REGEX, alias.split("_")[0])
            if regex_match is not None:
                name = regex_match["gate"]
                qubits_str = regex_match["qubits"]
                qubits = ast.literal_eval(qubits_str)
                if f"{name}({qubits_str})" in self.gates_settings.gate_names:
                    return self.gates_settings.get_gate(name=name, qubits=qubits)

        element = self.instruments.get_instrument(alias=alias)
        if element is None:
            element = self.instrument_controllers.get_instrument_controller(alias=alias)
        if element is None:
            element = self._get_bus_by_alias(alias=alias)
        if element is None:
            element = self.chip.get_node_from_alias(alias=alias)
        return element

    def get_ch_id_from_qubit_and_bus(self, alias: str, qubit_index: int) -> int | None:
        """Finds a sequencer id for a given qubit given a bus alias. This utility is added so that one can get a qrm's
        channel id easily in case the setup contains more than one qrm and / or there is not a one to one correspondance
        between sequencer id in the instrument and the qubit id. This one to one correspondance used to be the norm for
        5 qubit chips with non-RF QRM modules with 5 sequencers, each mapped to a qubit with the same numerical id as the
        sequencer.
        For QCMs it is also useful since the sequencer id is not always the same as the qubit id.

        Args:
            alias (str): bus alias
            qubit_index (int): qubit index
        Returns:
            int: sequencer id
        """
        bus = next((bus for bus in self._get_bus_by_qubit_index(qubit_index=qubit_index) if bus.alias == alias), None)
        if bus is None:
            raise ValueError(f"Could not find bus with alias {alias} for qubit {qubit_index}")
        if instrument := next(
            (
                instrument
                for instrument in bus.system_control.instruments
                if instrument.name in [InstrumentName.QBLOX_QRM, InstrumentName.QRMRF]
            ),
            None,
        ):
            return next(
                sequencer.identifier for sequencer in instrument.awg_sequencers if sequencer.qubit == qubit_index
            )
        # if the alias is not in the QRMs, it should be in the QCM
        instrument = next(
            instrument
            for instrument in bus.system_control.instruments
            if instrument.name in [InstrumentName.QBLOX_QCM, InstrumentName.QCMRF]
        )
        return next(
            (sequencer.identifier for sequencer in instrument.awg_sequencers if sequencer.chip_port_id == bus.port),
            None,
        )

    def _get_bus_by_qubit_index(self, qubit_index: int) -> tuple[Bus, Bus, Bus]:
        """Finds buses associated with the given qubit index.

        Args:
            qubit_index (int): Qubit index to get the buses from.

        Returns:
            tuple[:class:`Bus`, :class:`Bus`, :class:`Bus`]: Tuple of Bus objects containing the flux, control and readout buses of the given qubit.
        """
        flux_port = self.chip.get_port_from_qubit_idx(idx=qubit_index, line=Line.FLUX)
        control_port = self.chip.get_port_from_qubit_idx(idx=qubit_index, line=Line.DRIVE)
        readout_port = self.chip.get_port_from_qubit_idx(idx=qubit_index, line=Line.FEEDLINE_INPUT)
        flux_bus = self.buses.get(port=flux_port)
        control_bus = self.buses.get(port=control_port)
        readout_bus = self.buses.get(port=readout_port)
        return flux_bus, control_bus, readout_bus

    def _get_bus_by_alias(self, alias: str | None = None):
        """Gets buses given their alias.

        Args:
            alias (str | None, optional): Bus alias to identify it. Defaults to None.

        Returns:
            :class:`Bus`: Bus corresponding to the given alias. If none is found `None` is returned.

        """
        return next((bus for bus in self.buses if bus.alias == alias), None)

    def get_parameter(self, parameter: Parameter, alias: str, channel_id: int | None = None):
        """Get platform parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            alias (str): Alias of the bus where the parameter is set.
            channel_id (int, optional): ID of the channel we want to use to set the parameter. Defaults to None.
        """
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if alias == "platform" or regex_match is not None:
            return self.gates_settings.get_parameter(alias=alias, parameter=parameter, channel_id=channel_id)
        element = self.get_element(alias=alias)
        return element.get_parameter(parameter=parameter, channel_id=channel_id)

    def set_parameter(
        self,
        parameter: Parameter,
        value: float | str | bool,
        alias: str,
        channel_id: int | None = None,
    ):
        """Set a parameter for a platform element.

        If connected to an instrument, this function updates both the cache of the :class:`.Platform` object and the
        instrument's value. Otherwise, it only stores the value in the cache. Subsequent ``connect()`` + ``initial_setup()``
        will apply the cached values into the real instruments.

        If you use ``set_parameter`` + ``ql.save_platform()``, the saved runcard will include the new "set" value, even without
        an instrument connection, as the cache values of the :class:`.Platform` object are modified.

        Args:
            parameter (Parameter): Name of the parameter to change.
            value (float | str | bool): New value to set in the parameter.
            alias (str): Alias of the bus where the parameter is set.
            channel_id (int, optional): ID of the channel you want to use to set the parameter. Defaults to None.
        """
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if alias == "platform" or regex_match is not None:
            self.gates_settings.set_parameter(alias=alias, parameter=parameter, value=value, channel_id=channel_id)
            return
        element = self.get_element(alias=alias)
        element.set_parameter(parameter=parameter, value=value, channel_id=channel_id)

    def _load_instruments(self, instruments_dict: list[dict]) -> list[Instrument]:
        """Instantiates all instrument classes from their respective dictionaries.

        Args:
            instruments_dict (list[dict]): List of dictionaries containing the settings of each instrument.

        Returns:
            list[Instrument]: List of instantiated instrument classes.
        """
        instruments = []
        for instrument in instruments_dict:
            local_dict = deepcopy(instrument)
            instruments.append(InstrumentFactory.get(local_dict.pop(RUNCARD.NAME))(settings=local_dict))
        return instruments

    def _load_instrument_controllers(self, instrument_controllers_dict: list[dict]) -> list[InstrumentController]:
        """Instantiates all instrument controller classes from their respective dictionaries.

        Args:
            instrument_controllers_dict (list[dict]): List of dictionaries containing
            the settings of each instrument controller.

        Returns:
            list[InstrumentController]: List of instantiated instrument controller classes.
        """
        instrument_controllers = []
        for instrument_controller in instrument_controllers_dict:
            local_dict = deepcopy(instrument_controller)
            instrument_controllers.append(
                InstrumentControllerFactory.get(local_dict.pop(RUNCARD.NAME))(
                    settings=local_dict, loaded_instruments=self.instruments
                )
            )
        return instrument_controllers

    def to_dict(self):
        """Returns all platform information as a dictionary, called the :ref:`runcard <runcards>`. Used for the platform serialization.

        Returns:
            dict: Dictionary of the serialized platform
        """
        name_dict = {RUNCARD.NAME: self.name}
        device_id = {RUNCARD.DEVICE_ID: self.device_id}
        gates_settings_dict = {RUNCARD.GATES_SETTINGS: self.gates_settings.to_dict()}
        chip_dict = {RUNCARD.CHIP: self.chip.to_dict() if self.chip is not None else None}
        buses_dict = {RUNCARD.BUSES: self.buses.to_dict() if self.buses is not None else None}
        instrument_dict = {RUNCARD.INSTRUMENTS: self.instruments.to_dict() if self.instruments is not None else None}
        instrument_controllers_dict = {
            RUNCARD.INSTRUMENT_CONTROLLERS: self.instrument_controllers.to_dict()
            if self.instrument_controllers is not None
            else None,
        }

        return (
            name_dict
            | device_id
            | gates_settings_dict
            | chip_dict
            | buses_dict
            | instrument_dict
            | instrument_controllers_dict
        )

    def __str__(self) -> str:
        """String representation of the platform.

        Returns:
            str: Name of the platform.
        """
        return str(YAML().dump(self.to_dict(), io.BytesIO()))

    def execute_qprogram(
        self, qprogram: QProgram, bus_mapping: dict[str, str] | None = None, debug: bool = False
    ) -> QProgramResults:
        """Execute a QProgram using the platform instruments.

        Args:
            qprogram (QProgram): The QProgram to execute.

        Returns:
            dict[str, list[Result]]: A dictionary of measurement results. The keys correspond to the buses a measurement were performed upon, and the values are the list of measurement results in chronological order.
        """
        bus_aliases = {bus_mapping[bus] if bus_mapping and bus in bus_mapping else bus for bus in qprogram.buses}
        buses = [self._get_bus_by_alias(alias=bus_alias) for bus_alias in bus_aliases]
        instruments = {
            instrument
            for bus in buses
            for instrument in bus.system_control.instruments
            if isinstance(instrument, (QbloxModule, QuantumMachinesCluster))
        }
        if all(isinstance(instrument, QbloxModule) for instrument in instruments):
            return self._execute_qprogram_with_qblox(qprogram=qprogram, bus_mapping=bus_mapping, debug=debug)
        if all(isinstance(instrument, QuantumMachinesCluster) for instrument in instruments):
            if len(instruments) != 1:
                raise NotImplementedError(
                    "Executing QProgram in more than one Quantum Machines Cluster is not supported."
                )
            cluster: QuantumMachinesCluster = instruments.pop()  # type: ignore[assignment]
            return self._execute_qprogram_with_quantum_machines(
                cluster=cluster, qprogram=qprogram, bus_mapping=bus_mapping, debug=debug
            )
        raise NotImplementedError("Executing QProgram in a mixture of instruments is not supported.")

    def _execute_qprogram_with_qblox(
        self, qprogram: QProgram, bus_mapping: dict[str, str] | None = None, debug: bool = False
    ) -> QProgramResults:
        # Compile QProgram
        qblox_compiler = QbloxCompiler()
        sequences = qblox_compiler.compile(qprogram=qprogram, bus_mapping=bus_mapping)
        buses = {bus_alias: self._get_bus_by_alias(alias=bus_alias) for bus_alias in sequences}

        if debug:
            with open("debug_qblox_execution.txt", "w", encoding="utf-8") as sourceFile:
                for bus_alias in sequences:
                    print(f"Bus {bus_alias}:", file=sourceFile)
                    print(str(sequences[bus_alias]._program), file=sourceFile)  # pylint: disable=protected-access
                    print(file=sourceFile)

        # Upload sequences
        for bus_alias in sequences:
            sequence_hash = hash_qpy_sequence(sequence=sequences[bus_alias])
            if bus_alias not in self._qpy_sequence_cache or self._qpy_sequence_cache[bus_alias] != sequence_hash:
                buses[bus_alias].upload_qpysequence(qpysequence=sequences[bus_alias])
                self._qpy_sequence_cache[bus_alias] = sequence_hash
            # sync all rellevant sequences
            for instrument in buses[bus_alias].system_control.instruments:
                if isinstance(instrument, QbloxModule):
                    instrument.sync_by_port(buses[bus_alias].port)

        # Execute sequences
        for bus_alias in sequences:
            buses[bus_alias].run()

        # Acquire results
        results = QProgramResults()
        for bus_alias in buses:
            if isinstance(buses[bus_alias].system_control, ReadoutSystemControl):
                acquisitions = list(sequences[bus_alias].todict()["acquisitions"])
                bus_results = buses[bus_alias].acquire_qprogram_results(acquisitions=acquisitions)
                for bus_result in bus_results:
                    results.append_result(bus=bus_alias, result=bus_result)

        # Reset instrument settings
        for bus_alias in sequences:
            for instrument in buses[bus_alias].system_control.instruments:
                if isinstance(instrument, QbloxModule):
                    instrument.desync_by_port(buses[bus_alias].port)

        return results

    def _execute_qprogram_with_quantum_machines(  # pylint: disable=too-many-locals
        self,
        cluster: QuantumMachinesCluster,
        qprogram: QProgram,
        bus_mapping: dict[str, str] | None = None,
        debug: bool = False,
    ) -> QProgramResults:
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, measurements = compiler.compile(qprogram=qprogram, bus_mapping=bus_mapping)

        cluster.append_configuration(configuration=configuration)

        if debug:
            with open("debug_qm_execution.py", "w", encoding="utf-8") as sourceFile:
                print(generate_qua_script(qua_program, cluster.config), file=sourceFile)

        compiled_program_id = cluster.compile(program=qua_program)
        job = cluster.run_compiled_program(compiled_program_id=compiled_program_id)

        acquisitions = cluster.get_acquisitions(job=job)

        results = QProgramResults()
        for measurement in measurements:
            measurement_result = QuantumMachinesMeasurementResult(
                *[acquisitions[handle] for handle in measurement.result_handles]
            )
            results.append_result(bus=measurement.bus, result=measurement_result)

        return results

    def execute(
        self,
        program: PulseSchedule | Circuit,
        num_avg: int,
        repetition_duration: int,
        num_bins: int = 1,
        queue: Queue | None = None,
    ) -> Result | QbloxResult:
        """Compiles and executes a circuit or a pulse schedule, using the platform instruments.

        If the ``program`` argument is a :class:`Circuit`, it will first be translated into a :class:`PulseSchedule` using the transpilation
        settings of the platform. Then the pulse schedules will be compiled into the assembly programs and executed.

        To compile to assembly programs, the ``platform.compile()`` method is called; check its documentation for more information.

        Args:
            program (:class:`PulseSchedule` | :class:`Circuit`): Circuit or pulse schedule to execute.
            num_avg (int): Number of hardware averages used.
            repetition_duration (int): Minimum duration of a single execution.
            num_bins (int, optional): Number of bins used. Defaults to 1.
            queue (Queue, optional): External queue used for asynchronous data handling. Defaults to None.

        Returns:
            Result: Result obtained from the execution. This corresponds to a numpy array that depending on the
                platform configuration may contain the following:

                - Scope acquisition is enabled: An array with dimension `(2, N)` which contain the scope data for
                    path0 (I) and path1 (Q). N corresponds to the length of the scope measured.

                - Scope acquisition disabled: An array with dimension `(#sequencers, 2, #bins)`.
        """
        # Compile pulse schedule
        programs = self.compile(program, num_avg, repetition_duration, num_bins)

        # Upload pulse schedule
        for bus_alias in programs:
            bus = self._get_bus_by_alias(alias=bus_alias)
            bus.upload()

        # Execute pulse schedule
        for bus_alias in programs:
            bus = self._get_bus_by_alias(alias=bus_alias)
            bus.run()

        # Acquire results
        readout_buses = [
            bus for bus in self.buses if isinstance(bus.system_control, ReadoutSystemControl) and bus.alias in programs
        ]
        results: list[Result] = []
        for bus in readout_buses:
            result = bus.acquire_result()
            if queue is not None:
                queue.put_nowait(item=result)
            results.append(result)

        for instrument in self.instruments.elements:
            if isinstance(instrument, QbloxModule):
                instrument.desync_sequencers()

        # Flatten results if more than one readout bus was used for a qblox module
        if len(results) > 1:
            results = [
                QbloxResult(
                    integration_lengths=[length for result in results for length in result.integration_lengths],  # type: ignore [attr-defined]
                    qblox_raw_results=[raw_result for result in results for raw_result in result.qblox_raw_results],  # type: ignore [attr-defined]
                )
            ]
        if not results:
            raise ValueError("There are no readout buses in the platform.")

        if isinstance(program, Circuit):
            results = [self._order_result(results[0], program)]

        # FIXME: resurn result instead of results[0]
        return results[0]

    def _order_result(self, result: Result, circuit: Circuit) -> Result:
        """Order the results of the execution as they are ordered in the input circuit.
        Finds the absolute order of each measurement for each qubit and its corresponding key in the
        same format as in qblox's aqcuisitions dictionary (#qubit, #qubit_measurement).
        Then it orders results in the same measurement order as the one in circuit.queue.
        Args:
            result (Result): Result obtained from the execution
            circuit (Circuit): qibo circuit being executed
        Returns:
            Result: Result obtained from the execution, with each measurement in the same order as in circuit.queue
        """
        if not isinstance(result, QbloxResult):
            raise NotImplementedError("Result ordering is only implemented for qblox results")

        # register the overall order of all qubit measurements.
        qubits_m = {}
        order = {}
        # iterate over qubits measured in same order as they appear in the circuit
        for i, qubit in enumerate(qubit for gate in circuit.queue for qubit in gate.qubits if isinstance(gate, M)):
            if qubit not in qubits_m:
                qubits_m[qubit] = 0
            order[(qubit, qubits_m[qubit])] = i
            qubits_m[qubit] += 1
        if len(order) != len(result.qblox_raw_results):
            raise ValueError(
                f"Number of measurements in the circuit {len(order)} does not match number of acquisitions {len(result.qblox_raw_results)}"
            )

        # allocate each measurement its corresponding index in the results list
        results = [None] * len(order)  # type: list | list[dict]
        for qblox_result in result.qblox_raw_results:
            measurement = qblox_result["measurement"]
            qubit = qblox_result["qubit"]
            results[order[(qubit, measurement)]] = qblox_result

        return QbloxResult(integration_lengths=result.integration_lengths, qblox_raw_results=results)

    def compile(
        self, program: PulseSchedule | Circuit, num_avg: int, repetition_duration: int, num_bins: int
    ) -> dict[str, list[QpySequence]]:
        """Compiles the circuit / pulse schedule into a set of assembly programs, to be uploaded into the awg buses.

        If the ``program`` argument is a :class:`Circuit`, it will first be translated into a :class:`PulseSchedule` using the transpilation
        settings of the platform. Then the pulse schedules will be compiled into the assembly programs.

        This methods gets called during the ``platform.execute()`` method, check its documentation for more information.

        Args:
            program (:class:`PulseSchedule` | :class:`Circuit`): Circuit or pulse schedule to compile.
            num_avg (int): Number of hardware averages used.
            repetition_duration (int): Minimum duration of a single execution.
            num_bins (int): Number of bins used.

        Returns:
            dict: Dictionary of compiled assembly programs. The key is the bus alias (``str``), and the value is the assembly compilation (``list``).

        Raises:
            ValueError: raises value error if the circuit execution time is longer than ``repetition_duration`` for some qubit.
        """
        # We have a circular import because Platform uses CircuitToPulses and vice versa

        if isinstance(program, Circuit):
            transpiler = CircuitTranspiler(platform=self)
            pulse_schedule = transpiler.transpile_circuit(circuits=[program])[0]
        elif isinstance(program, PulseSchedule):
            pulse_schedule = program
        else:
            raise ValueError(
                f"Program to execute can only be either a single circuit or a pulse schedule. Got program of type {type(program)} instead"
            )
        return self.compiler.compile(
            pulse_schedule=pulse_schedule, num_avg=num_avg, repetition_duration=repetition_duration, num_bins=num_bins
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Rectangular pulse shape."""
from copy import deepcopy
from dataclasses import dataclass

import numpy as np
from scipy.special import erf

from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class FlatTop(PulseShape):
    """Rectangular/square pulse shape. Given by a constant height line.

    Examples:
        To get the envelope of a rectangular shape, with ``amplitude`` equal to ``X``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import Rectangular
            rectangular_envelope = Rectangular().envelope(amplitude=X, duration=50)

        which for ``X`` being ``1.`` and ``0.75``, look respectively like:

        .. image:: /classes_images/rectangulars.png
            :width: 800
            :align: center
    """

    name = PulseShapeName.FLATTOP  #: Name of the rectangular pulse shape.
    gaussian: float

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0, buffer: float = 3.0) -> np.ndarray:
        """Constant amplitude envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """
        x = np.arange(0, duration, resolution)
        A = amplitude
        g = self.gaussian
        buf = buffer
        dur = duration
        return 0.5 * A * np.real((erf(g * x - buf) - erf(g * (x - (dur + -buf / g)))))

    @classmethod
    def from_dict(cls, dictionary: dict) -> "FlatTop":
        """Loads Rectangular object/shape from dictionary.

        Args:
            dictionary (dict): Dictionary representation of the Rectangular object/shape including the name of the pulse shape.

        Returns:
            Rectangular: Loaded class.
        """
        local_dictionary = deepcopy(dictionary)
        local_dictionary.pop("name", None)
        return cls(**local_dictionary)

    def to_dict(self) -> dict:
        """Returns dictionary representation of the Rectangular object/shape.

        Returns:
            dict: Dictionary representation of the pulse shape including its name.
        """
        return {"name": self.name.value, "gaussian": self.gaussian}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Experiment Options Typings """
import io
from dataclasses import asdict, dataclass, field

from ruamel.yaml import YAML

from qililab.constants import EXPERIMENT, RUNCARD
from qililab.utils.loop import Loop

DEFAULT_EXPERIMENT_NAME = "experiment"


@dataclass
class ExperimentSettings:
    """Experiment settings."""

    hardware_average: int = 1024
    software_average: int = 1
    repetition_duration: int = 200000
    num_bins: int = 1

    def __str__(self):
        """Returns a string representation of the experiment settings."""
        return str(YAML().dump(asdict(self), io.BytesIO()))


@dataclass
class ExperimentOptions:
    """Experiment Options"""

    loops: list[Loop] | None = None
    settings: ExperimentSettings = field(default_factory=ExperimentSettings)
    name: str = DEFAULT_EXPERIMENT_NAME
    remote_save: bool = True
    description: str = ""

    def to_dict(self):
        """Convert Experiment into a dictionary.

        Returns:
            dict: Dictionary representation of the Experiment class.
        """
        return {
            EXPERIMENT.LOOPS: [loop.to_dict() for loop in self.loops] if self.loops is not None else None,
            RUNCARD.GATES_SETTINGS: asdict(self.settings),
            RUNCARD.NAME: self.name,
            EXPERIMENT.REMOTE_SAVE: self.remote_save,
            EXPERIMENT.DESCRIPTION: self.description,
        }

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Load Experiment Options from dictionary.

        Args:
            dictionary (dict): Dictionary description of an experiment.
        """
        input_loops = dictionary[EXPERIMENT.LOOPS]
        loops = [Loop(**loop) for loop in input_loops] if input_loops is not None else None

        return ExperimentOptions(
            loops=loops,
            settings=ExperimentSettings(**dictionary[RUNCARD.GATES_SETTINGS])
            if RUNCARD.GATES_SETTINGS in dictionary
            else ExperimentSettings(),
            name=dictionary[RUNCARD.NAME] if RUNCARD.NAME in dictionary else DEFAULT_EXPERIMENT_NAME,
            remote_save=dictionary.get(EXPERIMENT.REMOTE_SAVE, True),
            description=dictionary.get(EXPERIMENT.DESCRIPTION, ""),
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=anomalous-backslash-in-string, inconsistent-return-statements
"""Automatic-calibration Controller module, which works with notebooks as nodes."""
from datetime import datetime, timedelta

import networkx as nx
import pandas as pd

import qililab as ql
from qililab.calibration.calibration_node import CalibrationNode
from qililab.config import logger
from qililab.data_management import build_platform, save_platform
from qililab.platform.platform import Platform


class CalibrationController:
    """Controls the automatic calibration sequence.

    **Usage:**
        - To calibrate the full graph, use the ``run_automatic_calibration()`` method.
        - To target and ensure a node works, use the ``maintain(node)`` method.

    |

    **Graph Structure:**

    In the graph, directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration).
    This defines our `starts` and `ends` of the calibration:

    - `starts`: `Roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

    - `ends`: `Leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

    .. code-block:: python

        #                     /--> 2 -->\\
        #                    /     |     \\
        #    (start, root)  0      |      3 --> 4 (end, leave)
        #                    \     v     /
        #                     \--> 1 -->/

    .. note::

        Find information about the automatic-calibration workflow, in the examples below.

    Args:
        calibration_graph (nx.DiGraph): The calibration (directed acyclic) graph, where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``. Directions should be given
            by `nodes` pointing to their next `dependents` (natural time flow for calibration), defining our `starts` and `ends` of the calibration as the `roots` (``in_degree=0``) and `leaves`
            (``out_degree=0``) of the graph.

        node_sequence (dict[str, CalibrationNode]): Mapping for the nodes of the graph, from strings into the actual initialized nodes.
        runcard (str): The runcard path, containing the serialized platform where the experiments will be run.

    Examples:

        **Calibration Workflow:**

        The calibration process is structured into three levels of methods:

        1. **Highest Level Method**: The ``run_automatic_calibration()`` method finds all the end nodes of the graph (`leaves`, those without further `dependents`) and runs ``maintain()`` on them.

        2. **Mid-Level Methods**: ``maintain()`` and ``diagnose()``.
            - ``maintain(node)`` starts from the `roots` that ``node`` depends on, and moves forwards (`dependency -> dependant`) until ``node``, checking the last time executions and data at each step. If a problem (``bad_data``) is found, it calls ``diagnose()`` to solve it.
            - ``diagnose(node)`` does more strict checks, fixing inaccuracies in the system's state to allow ``maintain()`` to continue. It works in reverse, starting from the problematic (``bad_data``) node, it goes back (`dependency <- dependant`) until it finds the origin of the problem.

        3. **Low-Level Methods**: ``check_state()``, ``check_data()``, and ``calibrate()`` are the methods you would be calling during this process to interact with the ``nodes``.

        |

        Finally, ``run_automatic_calibration()`` is designed to start acquiring data or calibrating in the optimal location of the graph to avoid extra work:

        # TODO: Change this workflow explanations to new one
        - If node A has been calibrated very recently (before the ``drift_timeout`` of the :class:`.CalibrationNode`), it would be a waste of resources to check its data, so ``check_state()`` makes ``maintain()`` skip it.
        - If node A depends on node B, before calibrating node A, we check the data of node B. Calibrating A would be a waste of resources if we were doing so based on faulty data, so it goes to its dependencies first.

        .. note:: Find more information about the automatic calibration workflow at https://arxiv.org/abs/1803.03226.

        ----------

        **Dangerous Behaviors:**

        Note that depending on your ``CalibrationController`` construction, you can have dangerous behaviors in the workflow. You need to watch out for:

        # TODO: change this dangerous behaviours explanations, to new workflow
        - If you give bad ``comparison_thresholds`` or have bad ``comparison_models``, returning ``out_of_spec`` when you actually have ``bad_data`` or the other way around, will make ``diagnose()`` not being able to work properly, since for it to work, you need to have a single layer node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` nodes.
        - If you give too long ``drift_timeout``'s, since ``maintain()`` will assume the node is 100% working, you will go to further dependents when maybe you shouldn't without recalibrating. In the end ``diagnose()`` saves the day here, since it doesn't do ``check_state()``, although it will be less efficient.

        ----------

        **Practical example:**

        To create two linked nodes twice, for two different qubits, and pass them to a :class:`.CalibrationController` and run a ``maintain()``, you need:

        .. code-block:: python

            import numpy as np
            import networkx as nx

            from qililab.calibration import CalibrationController, CalibrationNode, norm_root_mean_sqrt_error

            # GRAPH CREATION AND NODE MAPPING (key = name in graph, value = node object):
            nodes = {}
            G = nx.DiGraph()
            first, second = [], []

            # CREATE NODES:
            for qubit in [0, 1]:
                first[qubit] = CalibrationNode(
                    nb_path="notebooks/first.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=4,
                    bad_data_threshold=8,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1800.0,
                )
                nodes[first[qubit].node_id] = first[qubit]

                second[qubit] = CalibrationNode(
                    nb_path="notebooks/second.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=2,
                    bad_data_threshold=4,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1.0,
                    sweep_interval=np.arange(start=0, stop=19, step=1),
                )
                nodes[second[qubit].node_id] = second[qubit]

                # GRAPH BUILDING (1 --> 2):
                G.add_edge(first[qubit].node_id, second[qubit].node_id)

            # CREATE CALIBRATION CONTROLLER:
            controller = CalibrationController(node_sequence=nodes, calibration_graph=G, runcard=path_runcard)

            ### WORKFLOW TO DO:
            controller.maintain(second[1]) # maintain second node for qubit 1

        .. note::

            Find information about how these nodes and their notebooks need to be in the :class:`CalibrationNode` class documentation.

            There you will also find the above code, but without defining ``first`` and ``second`` as lists.
    """

    def __init__(self, calibration_graph: nx.DiGraph, node_sequence: dict[str, CalibrationNode], runcard: str):
        if not nx.is_directed_acyclic_graph(calibration_graph):
            raise ValueError("The calibration graph must be a Directed Acyclic Graph (DAG).")

        self.calibration_graph: nx.DiGraph = calibration_graph
        """The calibration (directed acyclic) graph. Where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``.

        Directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration),
        defining our `starts` and `ends` of the calibration, as:

        - `starts`: `roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

        - `ends`: `leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

        .. code-block:: python

            #                     /--> 2 -->\\
            #                    /     |     \\
            #    (start, root)  0      |      3 --> 4 (end, leave)
            #                    \     v     /
            #                     \--> 1 -->/
        """

        self.node_sequence: dict[str, CalibrationNode] = node_sequence
        """Mapping for the nodes of the graph, from strings into the actual initialized nodes (dict)."""

        self.runcard: str = runcard
        """The runcard path, containing the serialized platform where the experiments will be run (str)."""

        self.platform: Platform = build_platform(runcard)
        """The initialized platform, where the experiments will be run (Platform)."""

    def calibrate_all(self, node: CalibrationNode):
        """Calibrates all the nodes sequentially.

        Args:
            node (CalibrationNode): The node where we want to start the `calibration_all()` on. Normally you would want
                this node to be the furthest node in the calibration graph.
        """
        logger.info("WORKFLOW: Calibrating all %s.\n", node.node_id)
        for n in self._dependencies(node):
            self.calibrate_all(n)

        # You can skip it from 2h time, but also skip it due to `been_calibrated()`
        # TODO: DOCUMENT: If you want to start the calibration from the start again, just remove the executed files!
        if (
            node.previous_timestamp is None or self._is_timeout_expired(node.previous_timestamp, 7200.0)
        ) and not node.been_calibrated:
            self.calibrate(node)
            self._update_parameters(node)

    def run_automatic_calibration(
        self,
        force_maintain: bool = False,
        safe_diagnose: bool = True,
        strict_dependencies: bool = False,
        run_fidelities: bool = False,
        calibrate_all: bool = False,
    ) -> dict[str, dict]:
        """Runs the full automatic calibration procedure and retrieves the final set parameters and achieved fidelities dictionaries.

        This is the primary interface for our calibration procedure and the highest level algorithm, which finds all the end nodes of the graph
        (`leaves`, those without further `dependents`) and runs ``maintain()`` on them.

        Args:
            force_maintain_timeout_ratio (float, optional): Argument needed to compute the force maintain condition. Defaults to 0.0.
            safe_diagnose (bool, optional): Flag to specify if we make sure to avoid corner cases while diagnosing (its slower). Defaults to False.
            strict_dependencies (bool, optional): Flag to specify if the drift_timeout dependencies should be completely strict, meaning that a previous
                dependency can never have a smaller drift_timeout. Defaults to False.

        Returns:
            dict[str, dict]: Dictionary for the last set parameters and the last achieved fidelities. It contains two dictionaries (dict[tuple, tuple]) in the keys:
                - "set_parameters": Set parameters dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``str``: bus alias, int: qubit).
                    - value: (``float``: parameter value, ``str``: ``node_id`` where computed, ``datetime``: updated time).

                - "fidelities": Fidelities dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``int``: qubit).
                    - value: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        highest_level_nodes = [
            self.node_sequence[node]
            for node, out_degree in self.calibration_graph.out_degree()
            if (out_degree == 0 and not self.node_sequence[node].fidelity)
        ]

        # TODO: Do a couple of integrations test, with deterministic scripts, where you know the sequence of check_datas and
        # TODO: calibrates the workflow would need to follow, like the ones in my notebook, and check them.

        for node in highest_level_nodes:
            if calibrate_all:
                self.calibrate_all(node)

            else:
                self.maintain(
                    node,
                    force_mantain=force_maintain,
                    safe_diagnose=safe_diagnose,
                    strict_dependencies=strict_dependencies,
                )

        if run_fidelities:
            self.run_fidelities()

        logger.info(
            "\n#############################################\n"
            "Automatic calibration completed successfully!\n"
            "#############################################\n"
        )
        return self.get_qubit_fidelities_and_parameters_df_tables()

    def get_qubit_fidelities_and_parameters_df_tables(self) -> dict[str, pd.DataFrame]:
        """Generates the 1q, 2q, fidelities and parameters dataframes, with the last calibrations.

        Returns:
            dict[str, pd.DataFrame]: Last calibrations dataframes.
        """
        df_1q, df_2q = self.get_qubits_tables()

        return {
            "1q_table": df_1q,
            "2q_table": df_2q,
            "set_parameters": self.get_last_set_parameters(),
            "fidelities": self.get_last_fidelities(),
        }

    def run_fidelities(self) -> None:
        """Runs the fidelities notebooks."""
        fidelities_nodes = (node for node in self.node_sequence.values() if node.fidelity)
        for node in fidelities_nodes:
            self.calibrate(node)

    def maintain(
        self,
        node: CalibrationNode,
        force_mantain: bool = False,
        safe_diagnose: bool = False,
        strict_dependencies: bool = False,
    ) -> None:
        """Calls all the necessary subroutines (``check_state()``, ``check_data()`` and ``calibrate()``) in the respective dependencies to get a node
        in spec. Maintain contains the main workflow for our calibration procedure, and it's what is called from ``run_automatic_calibration()`` into
        each of the end ``nodes`` (leaves) of our graph.

        The algorithm starts from the `roots` that the passed ``node`` depends on, and moves forwards (`dependency -> dependant`) up to ``node``,
        at each step, first efficiently checking the last time executions (``check_state()``) until some fails, and then passes to check the data
        (``check_data()``, less efficient). If a problematic node is found, meaning, it doesn't pass ``check_state()`` and gives
        ``bad_data`` in ``check_data()``, then ``maintain()`` calls ``diagnose()`` for solving it.

        During all this process, the mid-level methods would be calling these low-level methods, that act on the nodes:
        ``calibrate()``, ``check_status()`` and ``check_data()``.

        Finally, ``maintain`` is designed to start acquiring data or calibrating in the optimal location of the graph, to avoid extra work:

        # TODO: Change this workflow explanations to new one
        - if node A has been calibrated very recently (before the ``drift_timeout`` of the :class:`.CalibrationNode`), it would be waste of resources to check its data, so ``check_state()`` makes ``maintain()`` skip it.
        - if node A depends on node B, before calibrating node A we check the data of node B, calibrating A would be a waste of resources if we were doing so based on faulty data, so it goes to its dependencies first.

        |

        # TODO: change this dangerous behaviours explanations, to new workflow
        Note that due to this, we can have dangerous behaviours in ``maintain()``, for example:

        - if you give bad ``comparison_thresholds`` or have bad ``comparison_models``, returning ``out_of_spec`` when you actually have ``bad_data`` or the other way around, making ``diagnose()`` not being able to work properly. Since for ``diagnose()`` to work, you need to have a single layer node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` nodes.
        - if you give too long ``drift_timeout``'s, since ``maintain()`` will assume the node is 100% working, you will go to further dependants, when maybe you shouldn't without recalibrating (in the end ``diagnose()`` saves the day here, since it doesn't do ``check_state()``, but in a less efficient way).

        .. note:: Find more information about the ``maintain()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where we want to start the algorithm on, getting it in spec. Normally you would want
                this node to be the furthest node in the calibration graph.
            force_mantain (bool, optional): Flag to force the method to not considerate `check_status` output value,
                meaning it allways perform a call to `check_data`. Default to False.
            safe_diagnose (bool, optional): Flag to specify if we make sure to avoid corner cases while diagnosing (its slower). Defaults to False.
        """
        logger.info("WORKFLOW: Maintaining %s.\n", node.node_id)
        # Recursion over all the nodes that the current node depends on.
        for n in self._dependencies(node):
            self.maintain(n)

        # TODO: Integrationt test, where if all the drift_timeouts are super big, no check_data or calibration is done
        # TODO: when no force_maintain, and one in the end, when the flag exist.

        # If check_state of this node passes, don't check data, assume it works (unless its an end node).
        node_status = self.check_state(node, strict_dependencies=strict_dependencies)
        if not force_mantain and node_status:
            return

        # Check data, if bad, diagnose to find the problem.
        result = self.check_data(node)

        # If in_spec, maintain ends!
        if result == "in_spec":
            self._update_parameters(node)  # In case the runcard has changed from last calibration.
            return

        # TODO: Integration test, for the second condition here: for example, if we had check_data like
        # TODO: V - V - O - O then it would calibrate the last without calibrating the dependency...

        # If `bad_data` or `out_of_spec` in the last nodes with safe flag:
        if result == "bad_data" or (safe_diagnose and result == "out_of_spec"):
            for n in self._dependencies(node):
                logger.info("WORKFLOW: Diagnosing %s from maintain(%s).\n", n.node_id, node.node_id)
                self.diagnose(n, safe=safe_diagnose)

        # Implicit out_spec case (except above second condition with safe flag)
        if force_mantain and node_status:
            logger.info(
                "`force_maintain` in node %s, detected that `check_status` was passed, but node was not `in_spec', perhaps `drift_timeouts` should be updated.",
                node.node_id,
            )

        self.calibrate(node)
        self._update_parameters(node)

    def diagnose(self, node: CalibrationNode, safe: bool = False):
        """Checks the data of all the dependencies of a node, until it finds the root of the problem with their data.

        This is a method called by ``maintain()`` in the special case that its call of ``check_data()`` finds bad data.

        ``Diagnose()`` workflow works in reverse, starting from the problematic (``bad_data``) node, it goes back (`dependency <- dependant`)
        until it finds the origin of the problem (the first and only ``out_of_spec`` of that path, since the previous will be ``in_spec`` and the
        followings in ``bad_data``).

        ``Maintain()`` assumes that our knowledge of the state of the system matches the actual state of the system: if we knew a node would
        return bad data, we wouldn't bother ``calibrating`` it. The fact that ``check_data()`` returns ``bad_data`` means that that's not the
        case, our knowledge of the systems's state is inaccurate. That why ``diagnose()`` does more strict checks, fixing inaccuracies in our
        knowledge of the system's state, to allow ``maintain()`` to continue.

        #TODO: Explain safe_check, this as example:
        Diagnose called with flag `safe_check` will make sure `out_spec` nodes are detected by reliying on actuall information of dependant nodes.
        Otherwise the algorithm will rely on `out_spec` classification method provided by the user via comparison model chosen.

        Note that `check` will cover corner cases that may be caused by user choices in exchange for time complexity as the number of
        calls to `check_data` will be higher on average.

        # TODO: change this dangerous behaviours explanations, to new workflow
        Finally mention, two important thing to have in mind:

        - if you give bad ``comparison_thresholds`` or have bad ``comparison_models``, which return ``out_of_spec`` when you actually have ``bad_data`` or the other way around, it will make your full calibration fail. Since for ``diagnose()`` to work, you need to have a single node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` ones.
        - if you have wrong ``drift_timeout``, the algorithm will be slower, but in the end ``diagnose()`` saves the day, achieving the full calibration, since it doesn't do ``check_state()``.

        .. note:: Find more information about the ``diagnose()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where we want to start the algorithm.
            safe (bool, optional):  Flag to specify if we make sure to avoid corner cases (its slower). Defaults to False.

        Returns:
            bool: True is there have been recalibrations, False otherwise. The return value is only used by recursive calls.
        """
        logger.info("WORKFLOW: diagnosing %s.\n", node.node_id)

        if safe:
            # in spec case
            if self.check_data(node) == "in_spec":
                self._update_parameters(node)  # In case the runcard has changed from last calibration.
                return

            # bad_data/out_spec case
            for n in self._dependencies(node):
                self.diagnose(n, safe=True)

        else:
            result = self.check_data(node)

            # in spec case
            if result == "in_spec":
                self._update_parameters(node)  # In case the runcard has changed from last calibration.
                return False

            # bad data case
            recalibrated = []
            if result == "bad_data":
                recalibrated = [self.diagnose(n, safe=False) for n in self._dependencies(node)]
                logger.info("WORKFLOW: Dependencies diagnoses of %s: %s\n", node.node_id, str(recalibrated))
            # If not empty and only filled with False's (not any True).
            if recalibrated != [] and not any(recalibrated):
                return False

        # calibrate
        self.calibrate(node)
        self._update_parameters(node)
        return safe

    def check_state(self, node: CalibrationNode, strict_dependencies: bool = False) -> bool:
        """Checks if the node's drift timeouts have passed since the last calibration or data check.

        These timeouts represent how long it usually takes for the parameters to drift, specified by the user.

        Conditions for ``check_state()`` to pass:
            - The cal has had ``check_data()`` or ``calibrate()`` pass within the timeout period.
            - The cal has not failed ``calibrate()`` without resolution.
            - No dependencies have been recalibrated since the last time ``check_data()`` or ``calibrate()`` was run on this cal.
            - All dependencies pass ``check_state()``.

        .. note:: Find more information about the ``check_state()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node whose state needs to be checked.

        Returns:
            bool: True if the parameter's drift timeout has not yet expired, False otherwise.
        """
        # Check if something happened and the timestamp could not be setted properly
        if node.previous_timestamp is None:
            logger.info("WORKFLOW: check_state of %s: False.\n", node.node_id)
            return False

        # Get the list of the dependencies that have been calibrated before this node, all of them should be True
        for n in self._dependencies(node):
            # TODO: Integration test that cases where check_states are like: V - X - V - V only skip the first node
            # Keep the smaller drift timeout, from all its dependencies.
            if strict_dependencies and n.drift_timeout < node.drift_timeout:
                node.drift_timeout = n.drift_timeout
                logger.warning(
                    "The drift_timeout of node %s is bigger than its dependency %s, so the later will be used.",
                    node.node_id,
                    n.node_id,
                )
            # A dependency has been calibrated before this node
            if n.previous_timestamp >= node.previous_timestamp:
                logger.info("WORKFLOW: check_state of %s: False.\n", node.node_id)
                return False
            # A dependency is expired
            if self._is_timeout_expired(n.previous_timestamp, n.drift_timeout):
                logger.info("WORKFLOW: check_state of %s: False.\n", node.node_id)
                return False

        # If this node concretely passes check_state
        is_timeout_not_expired = not self._is_timeout_expired(node.previous_timestamp, node.drift_timeout)
        logger.info(
            "WORKFLOW: check_state of %s: %r.\n",
            node.node_id,
            is_timeout_not_expired,
        )
        return is_timeout_not_expired

    def check_data(self, node: CalibrationNode) -> str:
        """Checks if the parameters found in the last calibration are still valid, doing a reduced execution of the notebook.

        To do this, ``check_data()`` runs the experiment only in a few points, randomly chosen within the ``sweep_interval``,
        and compares the results with the data obtained in the same points when the experiment was last ``calibrate()`` with the
        entire ``sweep_interval``.

        The comparison is done with the model is indicated by the ``comparison_model`` attribute of :class:`.CalibrationNode`,
        which returns a value indicating how well the data fits the model.

        This comparison is then classified as ``in_spec`` (still valid), ``out_of_spec`` (drifted, but close) or ``bad_data``
        (noise/doesn't follow the desired fit, or no previous execution) given the ``in_spec_threshold`` and ``bad_data_threshold``
        attributes of :class:`.CalibrationNode`.

        .. note:: Find more information about the ``check_data()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node whose parameters need to be checked.

        Returns:
            str: The status of the data, depending on how the current results compare with the obtained during the last full calibration.

            Concretely, depending on the provided thresholds and comparison models, it will return:

            - ``in_spec`` if the results are still acceptable to use.
            - ``out_of_spec`` if the results have drifted are not acceptable enough, but they are close, and still follow the desired fit.
            - ``bad_data`` if the results don't follow the desired fit, or are noisy, which should happen when dependencies have drifted. Or also if there are no previous executions.
        """
        # pylint: disable=protected-access
        # 2 hour interval from last in_spec, for assuming is still good:
        if (node.previous_inspec is None or self._is_timeout_expired(node.previous_inspec, 7200.0)) and (
            node.previous_timestamp is None or self._is_timeout_expired(node.previous_timestamp, 7200.0)
        ):
            timestamp = node.run_node(check=True)
        else:
            logger.info('WORKFLOW: Using recent `in_spec`-`check_data` or calibration in node "%s".\n', node.node_id)
            return "in_spec"

        # Comparison and obtained parameters:
        comparison_outputs = node.previous_output_parameters
        obtained_outputs = node.output_parameters

        # If no previous result, return bad_data:
        comparison_result = "bad_data"

        # Do comparison and return the result:
        # (obtained_outputs should never be None, since we just run the notebook)
        if comparison_outputs is not None and obtained_outputs is not None:
            # Get comparison from last notebook and the new obtained parameters.
            compar_params = comparison_outputs["check_parameters"]
            obtain_params = obtained_outputs["check_parameters"]

            # TODO: Intergation test, that this returns the expected values, for very sided thresholds to
            # TODO: the right or to the left...

            comparison_number = self._obtain_comparison(node, obtain_params, compar_params)

            if comparison_number <= node.in_spec_threshold:
                comparison_result = "in_spec"
                node.previous_inspec = datetime.timestamp(datetime.now())

            elif comparison_number <= node.bad_data_threshold:
                comparison_result = "out_of_spec"

        # Do the necessary following changes:
        logger.info("WORKFLOW: check_data of %s: %s.\n", node.node_id, comparison_result)
        node._add_string_to_checked_nb_name(comparison_result, timestamp)  # add comparison result tag to the file name.
        node.output_parameters = node.previous_output_parameters
        return comparison_result

    def calibrate(self, node: CalibrationNode) -> None:
        """Runs a node's experiment on its default values of the ``sweep_interval``.

        This method is responsible for calibrating a node to bring it within spec. The calibration process is node-specific
        and depends on the notebook implementation and the calibration parameters.

        .. note:: Find more information about the ``calibrate()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where the calibration experiment is run.
        """
        logger.info('WORKFLOW: Calibrating node "%s".\n', node.node_id)
        node.previous_timestamp = node.run_node()
        node.been_calibrated = True
        node._add_string_to_checked_nb_name("calibrated", node.previous_timestamp)  # pylint: disable=protected-access
        # add _calibrated tag to the file name, which doesn't have a tag.

    def _update_parameters(self, node: CalibrationNode) -> None:
        """Updates the node parameters value in the platform, after a calibration.

        If the node does not have an associated parameter, or the parameter attribute of the node is None,
        this function does nothing.

        Args:
            node (CalibrationNode): The node which parameters need to be updated in the platform.
        """
        if node.output_parameters is not None and "platform_parameters" in node.output_parameters:
            for bus_alias, qubit, param_name, param_value in node.output_parameters["platform_parameters"]:
                logger.info(
                    "Platform updated with: (bus: %s, q: %s, %s, %s).", bus_alias, qubit, param_name, param_value
                )
                self.platform.set_parameter(
                    alias=bus_alias, parameter=ql.Parameter(param_name), value=param_value, channel_id=qubit
                )

            save_platform(self.runcard, self.platform)

    def get_last_set_parameters(self) -> pd.DataFrame:
        """Retrieves the last set parameters of the graph.

        Returns:
            dict[tuple, tuple]: Set parameters dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``str``: bus alias, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        parameters: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "platform_parameters" in node.output_parameters
            ):
                for bus, qubit, parameter, value in node.output_parameters["platform_parameters"]:
                    parameters[(parameter, bus, qubit)] = (
                        value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        df = pd.DataFrame.from_dict(parameters).transpose()
        if len(df.columns) == 3:
            df.columns = ["value", "node_id", "datetime"]
        if df.index.nlevels == 3:
            df.index.names = ["parameter", "bus", "qubit"]
        return df

    def get_last_fidelities(self) -> pd.DataFrame:
        """Retrieves the last updated fidelities of the graph.

        Returns:
            dict[tuple, tuple]: Fidelities dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        fidelities: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "fidelities" in node.output_parameters
            ):
                for qubit, fidelity, value in node.output_parameters["fidelities"]:
                    fidelities[(fidelity, qubit)] = (
                        value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        df = pd.DataFrame.from_dict(fidelities).transpose()
        if len(df.columns) == 3:
            df.columns = ["fidelity", "node_id", "datetime"]
        if df.index.nlevels == 2:
            df.index.names = ["fidelity", "qubit"]
        return df

    def get_qubits_tables(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Retrieves the last updated fidelities of the graph in the 1 qubit and 2 qubit tables.

        Returns:
            tuple[pd.DataFrame]: Tables where columns are each fidelity or parameter, and rows each qubit.
        """
        q1_df, q2_df = self._create_empty_dataframes()

        for node in self.node_sequence.values():
            qubit_list = node.node_id.split("_")
            qubit = "_".join(
                [i for i in qubit_list if any(char == "q" for char in i) and any(char.isdigit() for char in i)]
            ).replace("q", "-")[1:]

            if len(qubit) == 1:
                q1_df = self._fill_qubits_columns(node, qubit, q1_df)

            else:
                q2_df = self._fill_qubits_columns(node, qubit, q2_df)

        # Reorder fidelities to the front of the dataframe:
        q1_df = self._reorder_fidelities(q1_df)
        q2_df = self._reorder_fidelities(q2_df)

        return q1_df, q2_df

    def _create_empty_dataframes(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Creates the structure of the dataframe for the qubits table.

        Returns:
            pd.DataFrame: Empty df, where columns are each fidelity or parameter, and rows each qubit.
        """
        q1_idx, q1_col, q2_idx, q2_col = [], [], [], []  # type: ignore[var-annotated]

        for node in self.node_sequence.values():
            qubit_list = node.node_id.split("_")
            qubit = "_".join(
                [i for i in qubit_list if any(char == "q" for char in i) and any(char.isdigit() for char in i)]
            ).replace("q", "-")[1:]

            if len(qubit) == 1:
                q1_idx, q1_col = self._get_idx_and_columns(node, qubit, q1_idx, q1_col)

            else:
                q2_idx, q2_col = self._get_idx_and_columns(node, qubit, q2_idx, q2_col)

        df_q1 = pd.DataFrame("-", q1_idx, q1_col)
        df_q1.index.name = "qubit"

        df_q2 = pd.DataFrame("-", q2_idx, q2_col)
        df_q2.index.name = "qubit"

        return df_q1, df_q2

    @staticmethod
    def _get_idx_and_columns(node: CalibrationNode, qubit: str, idx: list, col: list) -> tuple[list, list]:
        """Gets the index and columns for creating a dataframe for 1q or 2q.

        Args:
            node (CalibrationNode): Node for which we are currently creating the dataframe idx and columns.
            qubit (str): qubit for which we are currently creating the dataframe idx and columns, corresponding to the node.
            idx (list): idx to be added.
            col (list): column to be added.

        Returns:
            tuple[list, list]: list of idx and columns for creating the dataframe.
        """
        if qubit not in idx:
            idx.append(qubit)

        if node.output_parameters is not None and node.previous_timestamp is not None:
            if "fidelities" in node.output_parameters:
                for _, fidelity, _ in node.output_parameters["fidelities"]:
                    if fidelity not in col:
                        col.append(fidelity)

            if "platform_parameters" in node.output_parameters:
                for bus, _, parameter, _ in node.output_parameters["platform_parameters"]:
                    bus_list = str(bus).split("_")
                    bus = "_".join([x for x in bus_list if not any(char.isdigit() for char in x)])

                    if f"{str(parameter)}_{bus}" not in col:
                        col.append(f"{str(parameter)}_{bus}")

        return idx, col

    @staticmethod
    def _fill_qubits_columns(node: CalibrationNode, qubit: str, df: pd.DataFrame) -> pd.DataFrame:
        """Fills the qubit tables columns with the fidelities or parameters values.

        Args:
            node (CalibrationNode): Node currently being filled.
            qubit (str): qubit being filled, corresponding to the node.
            df (pd.DataFrame): Dataframe to fill.

        Returns:
            pd.DataFrame: Filled dataframe.
        """
        if node.output_parameters is not None and node.previous_timestamp is not None:
            if "fidelities" in node.output_parameters:
                for _, fidelity, value in node.output_parameters["fidelities"]:
                    df[fidelity][qubit] = value

            if "platform_parameters" in node.output_parameters:
                for bus, _, parameter, value in node.output_parameters["platform_parameters"]:
                    bus_list = str(bus).split("_")
                    bus = "_".join([x for x in bus_list if not any(char.isdigit() for char in x)])
                    df[f"{str(parameter)}_{bus}"][qubit] = value

        return df

    @staticmethod
    def _reorder_fidelities(df: pd.DataFrame) -> pd.DataFrame:
        """It moves the fidelities columns to the front of the dataframe.

        Args:
            df (pd.DataFrame): Dataframe to reorder.

        Returns:
            pd.DataFrame: Reordered dataframe.
        """
        for column in df.columns:
            if "fidelity" in column:
                first_column = df.pop(column)
                df.insert(0, column, first_column)

        return df

    def _dependencies(self, node: CalibrationNode) -> list:
        """Finds the dependencies of a node.

        If in our graph we have `A -> B`, then node B depends on node A. Thus calling this method on B would return A.

        Args:
            node (CalibrationNode): The nodes for which the dependencies need to be retrieved.

        Returns:
            list: The nodes that the argument node depends on.
        """
        return [self.node_sequence[node_name] for node_name in self.calibration_graph.predecessors(node.node_id)]

    @staticmethod
    def _obtain_comparison(node: CalibrationNode, obtained: dict[str, list], comparison: dict[str, list]) -> float:
        """Returns the error, given the chosen method, between the comparison and obtained samples.

        Args:
            node (CalibrationNode): node from which the data comparison will be done.
            obtained (dict): obtained samples to compare.
            comparison (dict): previous samples to compare.

        Returns:
            float: difference/error between the two samples.
        """
        return node.comparison_model(obtained, comparison)

    @staticmethod
    def _is_timeout_expired(timestamp: float, timeout: float) -> bool:
        """Checks if the time passed since the ``timestamp`` is greater than the ``timeout`` duration.

        Args:
            timestamp (float): Timestamp from which the time should be checked, described in UNIX timestamp format.
            timeout (float): The timeout duration in seconds.

        Returns:
            bool: True if the ``timeout`` has expired, False otherwise.
        """
        # Calculate the time that should have passed (timestamp + timeout duration), convert them to datetime objects:
        timestamp_dt = datetime.fromtimestamp(timestamp)
        timeout_duration = timedelta(seconds=timeout)
        timeout_time = timestamp_dt + timeout_duration

        # Check if the current time is greater than the timeout time
        current_time = datetime.now()
        return current_time > timeout_time

    # TODO: @Isaac, I've commented this, because we won't use a ratio parameter, its too ugly and complex for a first iteration...
    # I would just delete this function, but if you really want, keep this one with a fixed number like 600, or a fixed small ratio like 0.01
    # Like it is now, its super beautiful and clear, two flags, nothing complicated, if hardware ask, we can always add it!

    # @staticmethod
    # def _get_forced_maintain_condition(node: CalibrationNode, ratio: float = 0.01) -> bool:
    #     """Method to return if a Calibration Node should be force maintained or not.
    #     The condition checks if the time transcurred from the last calibration is greater than a ratio of the drift timeout of the node.

    #     Args:
    #         node (CalibrationNode): Calibration Node to get the the forced maintain condition
    #         ratio(flat, optional): Ratio used for the condition. Default to 0.0.

    #     Returns:
    #         bool: Returns True if the condition is met. Otherwise returns False.
    #     """
    #     comp = node.drift_timeout * ratio if ratio != 0 else 600
    #     now = datetime.timestamp(datetime.now())
    #     return now - node.previous_timestamp > comp if node.previous_timestamp is not None else False



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LivePlot class."""
from collections.abc import Iterator
from itertools import count
from warnings import warn

import numpy as np
from qiboconnection.api import API

from qililab.typings.enums import LivePlotTypes
from qililab.utils.loop import Loop


class LivePlot:  # pylint: disable=too-many-instance-attributes
    """Class used to live plot experiment results.

    This class supports 1D and 2D plots. When running multiple sequences, the sequencer index will always be plotted
    in the x axis.

    When running loops in parallel, the plotted values will correspond to the loop with more nested loops. If all of
    them have the same amount of nested loops, the first one will be plotted.

    Args:
        connection (API): QiboConnection API object.
        num_schedules (int): Number of circuits/pulse schedules.
        title (str): Title of the plot.
        loops (list[Loop]): List of loops.
    """

    def __init__(self, connection: API, num_schedules: int, title: str = "", loops: list[Loop] | None = None):
        self.connection = connection
        self.num_schedules = num_schedules
        # We use the loop with more nested loops, or the first loop if all of them have the same amount of nested loops
        self.loop = None if loops is None else loops[np.argmax([loop.num_loops for loop in loops])]
        self.plot_dim = max(loop.num_loops for loop in loops) if loops is not None else 0
        self.plot_dim += 1 if num_schedules > 1 else 0
        if self.plot_dim not in {1, 2}:
            warn(
                message=f"The experiment contains {self.plot_dim} dimensions. Live plotting only supports 1D and 2D plots."
                " The remaining dimensions won't be plotted.",
                category=UserWarning,
            )
        self.x_values, self.y_values = self._axis_values()
        self.x_iterator, self.y_iterator = self._iterator_values()
        self.plot_id = self.create_live_plot(title=title)

    def _axis_values(self) -> tuple[list, list | None]:
        """Generate the values for the X and Y axes.

        When plotting more than one sequence, the X axis will correspond to the sequence index. Otherwise, the X axis
        will correspond to the range of the first (outer) loop.

        If there is only one loop, the Y axis will be None. Otherwise, the Y axis will correspond to the range of the
        second (inner) loop.

        Returns:
            tuple[list, list | None]: Values for the X and Y axes respectively.
        """
        if self.loop is None:
            return list(range(self.num_schedules)), None
        if self.loop.loop is None:
            if self.num_schedules == 1:
                return list(self.loop.values), None
            return list(range(self.num_schedules)), list(self.loop.values)
        return list(self.loop.loop.values), list(self.loop.values)

    def _iterator_values(self) -> tuple[Iterator, Iterator]:
        """Returns the iterators used to send data to the plot.

        Returns:
            tuple[Iterator, Iterator]: Iterators for the X and Y axis respectively.
        """
        if self.y_values is not None:
            if self.loop is not None and self.loop.num_loops >= 2:
                ranges = np.meshgrid(self.x_values, self.y_values)
                return iter(ranges[0].flatten()), iter(ranges[1].flatten())
            return iter(self.x_values), iter(self.y_values)
        return iter(self.x_values), count()

    def create_live_plot(self, title: str) -> int:
        """Create live plot.

        Args:
            title (str): Title of the plot.
        """
        return self.connection.create_liveplot(
            title=title,
            x_label=self.x_label,
            y_label=self.y_label,
            z_label=self.z_label,
            plot_type=self.plot_type.value,
            x_axis=np.array(self.x_values),
            y_axis=np.array(self.y_values) if self.y_values is not None else None,
        )

    def send_points(self, value: float):
        """Send points to the live plot.

        This function gathers the X and Y coordinates from the iterators generated in the class constructor.

        Args:
            value (float): Value to send to the plot.
        """
        if self.y_values is None:
            x_value = next(self.x_iterator)
            self.connection.send_plot_points(plot_id=self.plot_id, x=float(x_value), y=value)
            return
        x_value = next(self.x_iterator)
        y_value = next(self.y_iterator)
        self.connection.send_plot_points(
            plot_id=self.plot_id,
            x=float(x_value),
            y=float(y_value),
            z=value,
        )

    @property
    def plot_type(self) -> LivePlotTypes:
        """Return plot type.

        Returns:
            LivePlotTypes: Type of the LivePlot
        """
        return LivePlotTypes.SCATTER if self.y_values is None else LivePlotTypes.HEATMAP

    @property
    def x_label(self) -> str:
        """Create x label for live plotting.

        Returns:
            str: X label.
        """
        if self.loop is None:
            return "Sequence idx"
        if self.loop.loop is None:
            return self.label(loop=self.loop)
        return self.label(loop=self.loop.loop)

    @property
    def y_label(self):
        """Create y label for live plotting.

        Returns:
            str: Y label.
        """
        if self.loop is None:
            return "Amplitude"
        if self.loop.loop is None:
            return "Amplitude"
        return self.label(loop=self.loop)

    @property
    def z_label(self) -> str | None:
        """Create z label for live plotting.

        Returns:
            str: Z label.
        """
        return "Amplitude" if self.plot_dim == 2 else None

    def label(self, loop: Loop) -> str:
        """Return plot label from loop object.

        Args:
            loop (~utils.loop.Loop): Loop class.

        Returns:
            str: Plot label.
        """
        instrument_name = loop.alias
        return f"{instrument_name}: {loop.parameter.value} "



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Driver for the Drive Bus class."""
from qililab.drivers.interfaces import AWG, CurrentSource, VoltageSource
from qililab.platform.components.bus_driver import BusDriver
from qililab.platform.components.bus_factory import BusFactory


@BusFactory.register
class FluxBus(BusDriver):
    """Qililab's driver for Flux Bus

    Args:
        alias: Bus alias.
        port (int): Port to target.
        awg (AWG): Sequencer.
        source (CurrentSource | VoltageSource): Bus source instrument.
        distortions (list): Distortions to apply in this Bus.

    Returns:
        BusDriver: BusDriver instance of type flux bus.
    """

    def __init__(
        self, alias: str, port: int, awg: AWG | None, source: CurrentSource | VoltageSource | None, distortions: list
    ):
        super().__init__(alias=alias, port=port, awg=awg, distortions=distortions)
        self.instruments["source"] = source



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Castings for Enums """
from dataclasses import fields
from enum import Enum


def cast_enum_fields(obj):  # pylint: disable=missing-function-docstring
    for field in fields(obj):
        if isinstance(field.type, type) and issubclass(field.type, Enum):
            setattr(obj, field.name, field.type(getattr(obj, field.name)))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from qpysequence import Acquisitions, Program
from qpysequence import Sequence as QpySequence
from qpysequence import Waveforms, Weights
from qpysequence.library import long_wait
from qpysequence.program import Block, Loop, Register
from qpysequence.program.instructions import Acquire, AcquireWeighed, Move, Play, ResetPh, SetAwgGain, SetPh, Stop, Wait
from qpysequence.utils.constants import AWG_MAX_GAIN, INST_MAX_WAIT

from qililab.config import logger
from qililab.instruments.awg_settings import AWGQbloxADCSequencer, AWGQbloxSequencer
from qililab.instruments.qblox import QbloxModule
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.pulse.pulse_event import PulseEvent
from qililab.pulse.pulse_schedule import PulseSchedule
from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import InstrumentName


class QbloxCompiler:  # pylint: disable=too-many-locals
    """Qblox compiler for pulse schedules. Its only public method is `compile`, which compiles a pulse schedule to qpysequences (see docs for `QBloxCompiler.compile`).
    The class object is meant to be initialized once, with `compile` running as many times as necessary. This way the class attributes do not have to be initialized
    at each single compilation.

    Args:
        platform (Platform): Platform object representing the laboratory setup used to control quantum devices.
    Raises:
        ValueError: at init if no readout module (QRM) is found in platform.
    """

    def __init__(self, platform):
        self.qblox_modules = [
            instrument for instrument in platform.instruments.elements if isinstance(instrument, QbloxModule)
        ]
        self.buses = platform.buses
        # init variables as empty
        self.nshots = None
        self.num_bins = None
        self.repetition_duration = None
        self.readout_modules = [InstrumentName.QBLOX_QRM, InstrumentName.QRMRF]
        self.control_modules = [InstrumentName.QBLOX_QCM, InstrumentName.QCMRF]

        if all(
            qblox.name not in self.readout_modules for qblox in self.qblox_modules
        ):  # Raise error if qrm is not found
            raise ValueError("No QRM modules found in platform instruments")

    def compile(
        self, pulse_schedule: PulseSchedule, num_avg: int, repetition_duration: int, num_bins: int
    ) -> dict[str, list[QpySequence]]:
        """Compiles a given pulse schedule to a Sequence (QpySequence). If the compiler object is created and not
        reinitialized between executions, previously run Sequences and PulseBusSchedules are saved to the corresponding's
        qblox_module sequences and cache, respectively. They can then be loaded from the stored data instead of being
        compiled again.

        Args:
            pulse_schedule (PulseSchedule): PulseSchedule to be executed
            num_avg (int): hardware average - number of times to average execution over
            repetition_duration (int): fixed execution duration. Execution always lasts `repetition_duration` ns. If execution is shorter than that, a wait time of `repetition_duration` - circuit_execution_time is added (this is used for qubti relaxation)
            num_bins (int): number of shots to execute

        Raises:
            ValueError: if circuit execution time is longer than `repetition_duration`
            NotImplementedError: if more than one readout module is found connected to the platform

        Returns:
            dict[str, list[QpySequence]]: dictionary where the keys correspond to the bus which has to execute a list of QpySequences, given in the values
        """
        if num_avg != self.nshots or repetition_duration != self.repetition_duration or num_bins != self.num_bins:
            self.nshots = num_avg
            self.repetition_duration = repetition_duration
            self.num_bins = num_bins
            for qblox_module in self.qblox_modules:
                qblox_module.clear_cache()

        sequencer_qrm_bus_schedules, sequencer_qcm_bus_schedules = self.get_sequencer_schedule(
            pulse_schedule, self.qblox_modules
        )

        compiled_sequences = {}  # type: dict[str, list[QpySequence]]

        # generally a sequencer_schedule is the schedule sent to a specific bus, except for readout,
        # where multiple schedules for different sequencers are sent to the same bus
        for sequencer, sequencer_schedule in sequencer_qrm_bus_schedules + sequencer_qcm_bus_schedules:
            # check if circuit lasts longer than repetition duration
            end_time = None if len(sequencer_schedule.timeline) == 0 else sequencer_schedule.timeline[-1].end_time
            if end_time is not None and end_time > self.repetition_duration:
                raise ValueError(
                    f"Circuit execution time cannnot be longer than repetition duration but found circuit time {end_time } > {repetition_duration} for qubit {sequencer_schedule.qubit}"
                )
            # get qblox module for this sequencer
            qblox_module = self._get_instrument_from_sequencer(sequencer)
            bus_alias = self.buses.get(sequencer_schedule.port).alias
            # create empty list if key does not exist
            if bus_alias not in compiled_sequences:
                compiled_sequences[bus_alias] = []
            # if the schedule is already compiled get it from the module's sequences
            if sequencer_schedule == qblox_module.cache.get(
                sequencer.identifier
            ):  # if it's already cached then dont compile
                compiled_sequences[bus_alias].append(qblox_module.sequences[sequencer.identifier])
                # If the schedule is in the cache, delete the acquisition data (if uploaded) # FIXME: acquisitions should be deleted after acquisitions and not at compilation
                if qblox_module.name in self.readout_modules and hasattr(
                    qblox_module, "device"
                ):  # TODO: remove hasattr when the fixme above is done
                    qblox_module.device.delete_acquisition_data(sequencer=sequencer.identifier, all=True)

            else:
                # compile the sequences
                sequence = self._translate_pulse_bus_schedule(sequencer_schedule, sequencer)
                compiled_sequences[bus_alias].append(sequence)
                qblox_module.cache[sequencer.identifier] = sequencer_schedule
                qblox_module.sequences[sequencer.identifier] = sequence

        # check for sequences which where not in the PulseSchedule but are uploaded to a given qrm and erase them from cache
        # this is only needed for the qrm since it has multiple sequencers for the same bus
        for qrm in self.qblox_modules:
            # skip if not qrm
            if qrm.name not in self.readout_modules:
                continue
            compiled_seqs = {
                sequencer.identifier for sequencer, _ in sequencer_qrm_bus_schedules if sequencer in qrm.awg_sequencers
            }
            # pop from cache if a sequencer id is cached in the qrm but not present in compiled sequences for that qrm
            for seq_id in [cached_seq_id for cached_seq_id in qrm.cache.keys() if cached_seq_id not in compiled_seqs]:
                _ = qrm.cache.pop(seq_id)
                _ = qrm.sequences.pop(seq_id)

        return compiled_sequences

    def _translate_pulse_bus_schedule(
        self, pulse_bus_schedule: PulseBusSchedule, sequencer: AWGQbloxSequencer
    ) -> QpySequence:
        """Translate a pulse sequence into a Q1ASM program, a waveform dictionary and
        acquisitions dictionary (that is, a QpySequence sequence).

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse bus schedule to translate.
            sequencer (AWGQbloxSequencer): sequencer to generate the program

        Returns:
            Sequence: Qblox Sequence object containing the program and waveforms.
        """
        waveforms = self._generate_waveforms(pulse_bus_schedule=pulse_bus_schedule)
        acquisitions = self._generate_acquisitions(sequencer, timeline=pulse_bus_schedule.timeline)
        program = self._generate_program(
            pulse_bus_schedule=pulse_bus_schedule, waveforms=waveforms, sequencer=sequencer
        )
        weights = self._generate_weights(sequencer=sequencer)  # type: ignore
        return QpySequence(program=program, waveforms=waveforms, acquisitions=acquisitions, weights=weights)

    def _generate_waveforms(self, pulse_bus_schedule: PulseBusSchedule):
        """Generate I and Q waveforms from a PulseSequence object.
        Args:
            pulse_bus_schedule (PulseBusSchedule): PulseSequence object.
        Returns:
            Waveforms: Waveforms object containing the generated waveforms.
        """
        waveforms = Waveforms()

        unique_pulses: list[tuple[int, PulseShape]] = []

        for pulse_event in pulse_bus_schedule.timeline:
            if (pulse_event.duration, pulse_event.pulse.pulse_shape) not in unique_pulses:
                unique_pulses.append((pulse_event.duration, pulse_event.pulse.pulse_shape))
                amp = pulse_event.pulse.amplitude
                sign = 1 if amp >= 0 else -1
                envelope = pulse_event.envelope(amplitude=sign * 1.0)
                real = np.real(envelope)
                imag = np.imag(envelope)
                pair = (real, imag)
                waveforms.add_pair(pair=pair, name=pulse_event.pulse.label())

        return waveforms

    def _generate_acquisitions(self, sequencer: AWGQbloxSequencer, timeline: list[PulseEvent]) -> Acquisitions:
        """Generate Acquisitions object, currently containing a single acquisition named "default", with num_bins = 1
        and index = 0.

        Args:
            sequencer (AWGQbloxSequencer): sequencer to which we generate the acquisitions dictionary
            timeline (list[PulseEvent]): time ordered list of pulse events to play

        Returns:
            Acquisitions: Acquisitions object.
        """
        # FIXME: is it really necessary to generate acquisitions for a QCM??
        acquisitions = Acquisitions()
        if self._get_instrument_from_sequencer(sequencer).name in self.control_modules:
            return acquisitions
        for i, pulse in enumerate(timeline):
            acquisitions.add(name=f"acq_q{pulse.qubit}_{i}", num_bins=self.num_bins, index=i)
        return acquisitions

    def _generate_program(  # pylint: disable=too-many-locals
        self, pulse_bus_schedule: PulseBusSchedule, waveforms: Waveforms, sequencer: AWGQbloxSequencer
    ) -> Program:
        """Generate Q1ASM program

        Args:
            pulse_sequence (PulseSequence): pulse sequence
            waveforms (Waveforms): waveforms
            sequencer (int): index of the sequencer to generate the program

        Returns:
            Program: Q1ASM program.
        """
        # get qblox module from sequencer
        qblox_module = self._get_instrument_from_sequencer(sequencer)
        MIN_WAIT = qblox_module._MIN_WAIT_TIME  # pylint: disable=protected-access

        # Define program's blocks
        program = Program()
        start = Block(name="start")
        start.append_component(ResetPh())
        program.append_block(block=start)
        # Create registers with 0 and 1 (necessary for qblox)
        weight_registers = Register(), Register()
        if qblox_module.name in self.readout_modules:
            self._init_weights_registers(registers=weight_registers, program=program)
        avg_loop = Loop(name="average", begin=int(self.nshots))  # type: ignore
        bin_loop = Loop(name="bin", begin=0, end=self.num_bins, step=1)
        avg_loop.append_component(bin_loop)
        program.append_block(avg_loop)
        stop = Block(name="stop")
        stop.append_component(Stop())
        program.append_block(block=stop)
        timeline = pulse_bus_schedule.timeline
        if len(timeline) > 0 and timeline[0].start_time != 0:
            bin_loop.append_component(long_wait(wait_time=int(timeline[0].start_time)))

        for i, pulse_event in enumerate(timeline):
            waveform_pair = waveforms.find_pair_by_name(pulse_event.pulse.label())
            wait_time = timeline[i + 1].start_time - pulse_event.start_time if (i < (len(timeline) - 1)) else MIN_WAIT

            # Allow wait times longer than INST_MAX_WAIT
            long_wait_time = 0
            if wait_time > INST_MAX_WAIT:
                long_wait_time = wait_time - 2 * MIN_WAIT
                wait_time = 2 * MIN_WAIT
            # Add instrucitons to play the pulse
            phase = int((pulse_event.pulse.phase % (2 * np.pi)) * 1e9 / (2 * np.pi))
            gain = int(np.abs(pulse_event.pulse.amplitude) * AWG_MAX_GAIN)  # np.abs() needed for negative pulses
            bin_loop.append_component(SetAwgGain(gain_0=gain, gain_1=gain))
            bin_loop.append_component(SetPh(phase=phase))
            bin_loop.append_component(
                Play(
                    waveform_0=waveform_pair.waveform_i.index,
                    waveform_1=waveform_pair.waveform_q.index,
                    # wait until next pulse if QCM. If QRM wait min time (4) and wait time is added after acquiring
                    wait_time=int(wait_time)
                    if qblox_module.name not in self.readout_modules
                    else MIN_WAIT,  # TODO: add time of flight
                )
            )
            if qblox_module.name in self.readout_modules:
                self._append_acquire_instruction(
                    loop=bin_loop,
                    bin_index=bin_loop.counter_register,
                    acq_index=i,
                    sequencer=sequencer,  # type: ignore
                    weight_regs=weight_registers,
                    wait=wait_time - MIN_WAIT if (i < len(timeline) - 1) else MIN_WAIT,
                )

            # Add long wait for wait time if necessary
            if long_wait_time != 0:
                bin_loop.append_component(long_wait(long_wait_time))

        if self.repetition_duration is not None:
            wait_time = self.repetition_duration - bin_loop.duration_iter
            if wait_time > qblox_module._MIN_WAIT_TIME:  # pylint: disable=protected-access
                bin_loop.append_component(long_wait(wait_time=wait_time))

        logger.info("Q1ASM program: \n %s", repr(program))  # pylint: disable=protected-access
        return program

    def _generate_weights(self, sequencer: AWGQbloxADCSequencer) -> Weights:  # type: ignore
        """Generate acquisition weights.

        Returns:
            Weights: Acquisition weights.
        """
        weights = Weights()

        if self._get_instrument_from_sequencer(sequencer).name in self.control_modules:
            return weights
        pair = ([float(w) for w in sequencer.weights_i], [float(w) for w in sequencer.weights_q])
        weights.add_pair(pair=pair, indices=(0, 1))
        return weights

    def _append_acquire_instruction(
        self,
        loop: Loop,
        bin_index: Register | int,
        acq_index: int,
        sequencer: AWGQbloxADCSequencer,
        weight_regs: tuple[Register, Register],
        wait: int,
    ):
        """Append an acquire instruction to the loop."""
        weighed_acq = sequencer.weighed_acq_enabled

        acq_instruction = (
            AcquireWeighed(
                acq_index=acq_index,
                bin_index=bin_index,
                weight_index_0=weight_regs[0],
                weight_index_1=weight_regs[1],
                wait_time=wait,
            )
            if weighed_acq
            else Acquire(
                acq_index=acq_index,
                bin_index=bin_index,
                wait_time=wait,
            )
        )
        # TODO: implement time of flight properly to remove this hack
        loop.append_component(Wait(wait_time=220))
        loop.append_component(acq_instruction)

    def _init_weights_registers(self, registers: tuple[Register, Register], program: Program):
        """Initialize the weights `registers` and place the required instructions in the setup block of the `program`."""
        move_0 = Move(0, registers[0])
        move_1 = Move(1, registers[1])
        setup_block = program.get_block(name="setup")
        setup_block.append_components([move_0, move_1], bot_position=1)

    def get_sequencer_schedule(
        self, pulse_schedule: PulseSchedule, qblox_instruments: list[QbloxModule]
    ) -> tuple[list[tuple[AWGQbloxSequencer, PulseBusSchedule]], list[tuple[AWGQbloxSequencer, PulseBusSchedule]]]:
        """This method returns a dictionary containing the pulse schedule to be sent to each sequencer.
        This corresponds to a PulseBusSchedule object. Note that for multiplexed QRM more than one PulseBusSchedule is sent to the same
        bus (feedline) but to different sequencers (as qubit schedules), which is why QRM and QCM pulses are handled asymetrically
        in the methods below - and also why it is not possible to just group QRM and QCM pulses together under a single dictionary.

        Args:
            pulse_schedule (PulseSchedule): pulse schedule to be executed
            qblox_instruments (list[QbloxModule]): list of connected qblox modules

        Raises:
            ValueError: raises an error if readoud pulses are targeted at more tan one port

        Returns:
            dict[AWGQbloxSequencer, PulseBusSchedule]: dictionary of the pulse schedule (dict value) corresponding to a
            given sequencer (dict key)
        """
        qrm_sequencers = [
            sequencer
            for instrument in qblox_instruments
            for sequencer in instrument.awg_sequencers
            if instrument.name in self.readout_modules
        ]
        feedline_ports = {sequencer.chip_port_id for sequencer in qrm_sequencers}
        qcm_sequencers = [
            sequencer
            for instrument in qblox_instruments
            for sequencer in instrument.awg_sequencers
            if instrument.name in self.control_modules
        ]

        control_pulses = [
            pulse_bus_schedule
            for pulse_bus_schedule in pulse_schedule.elements
            if pulse_bus_schedule.port not in feedline_ports
        ]
        readout_pulses = [
            pulse_bus_schedule
            for pulse_bus_schedule in pulse_schedule.elements
            if pulse_bus_schedule.port in feedline_ports
        ]

        qcm_bus_schedules = [
            (sequencer, schedule)
            for sequencer in qcm_sequencers
            for schedule in control_pulses
            if sequencer.chip_port_id == schedule.port
        ]

        # Readout pulses should all be in one bus pulse schedule since they are all in the same line (feedline_input), so we separate them qubit-wise
        qrm_bus_schedules = [
            (sequencer, qubit_schedule)
            for sequencer in qrm_sequencers
            for schedule in readout_pulses
            for qubit_schedule in schedule.qubit_schedules()
            if sequencer.qubit == qubit_schedule.qubit
        ]
        return qrm_bus_schedules, qcm_bus_schedules

    def _get_instrument_from_sequencer(self, sequencer: AWGQbloxSequencer) -> QbloxModule:
        """Get the qblox module to which a given sequencer belongs

        Args:
            sequencer (AWGQbloxSequencer): qblox sequencer

        Returns:
            QbloxModule: qblox module
        """
        return next(
            instrument for instrument in self.qblox_modules for seq in instrument.awg_sequencers if seq == sequencer
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ExecutionManager class."""
import matplotlib.pyplot as plt
import numpy as np

from qililab.system_control import ReadoutSystemControl
from qililab.utils import Waveforms

from .bus_execution import BusExecution


# TODO: Rename to Drawer and simplify class
class ExecutionManager:  # pylint: disable=too-few-public-methods
    """ExecutionManager class.

    This class only contains drawing capabilities. The `ExecutionManager` name is maintained for backwards
    compatibility.

    At init checks that the number of schedules matches all the schedules for each bus.

    Args:
        num_schedules (int): Number of pulse schedules used in the experiment.
        buses (list[BusExecution] | None): List of buses used in the experiment.
    """

    def __init__(self, num_schedules: int, buses: list[BusExecution] | None = None):
        self.num_schedules = num_schedules
        self.buses = buses if buses is not None else []
        for bus in self.buses:
            if len(bus.pulse_bus_schedules) != self.num_schedules:
                raise ValueError(
                    f"Error: number of schedules: {self.num_schedules} does not match "
                    + f"the length of the schedules in a bus: {bus.pulse_bus_schedules}"
                )

    def draw(  # pylint: disable=too-many-locals
        self,
        real: bool = True,
        imag: bool = True,
        absolute: bool = False,
        modulation: bool = True,
        linestyle: str = "-",
        resolution: float = 1.0,
        idx: int = 0,
    ):
        """Save figure with the waveforms/envelopes sent to each bus.

        You can plot any combination of the real (blue), imaginary (orange) and absolute (green) parts of the function.

        Args:
            Args:
            real (bool): True to plot the real part of the function, False otherwise. Default to True.
            imag (bool): True to plot the imaginary part of the function, False otherwise. Default to True.
            absolute (bool): True to plot the absolute of the function, False otherwise. Default to False.
            modulation (bool): True to plot the modulated wave form, False for only envelope. Default to True.
            linestyle (str): lineplot ("-", "--", ":"), point plot (".", "o", "x") or any other linestyle matplotlib accepts. Defaults to "-".
            resolution (float, optional): The resolution of the pulses in ns. Defaults to 1.0.

        Returns:
            Figure: Matplotlib figure with the waveforms sent to each bus.
        """
        figure, axes = plt.subplots(nrows=len(self.buses), ncols=1, sharex=True)
        plt.ylabel("Amplitude")
        figure.suptitle(" Pulses on each bus ", fontsize=20)

        if len(self.buses) == 1:
            axes = [axes]  # make axes subscriptable

        for axis_idx, (bus_alias, waveforms) in enumerate(
            self._waveforms_dict(modulation=modulation, resolution=resolution, idx=idx).items()
        ):
            time = np.arange(len(waveforms)) * resolution
            axes[axis_idx].set_title(f"Bus {bus_alias}", loc="right")

            if imag:
                axes[axis_idx].plot(time, waveforms.q, linestyle, label="imag", color="orange")

            if real:
                axes[axis_idx].plot(time, waveforms.i, linestyle, label="real", color="blue")

            if absolute:
                waveform_abs = np.sqrt(waveforms.q * waveforms.q + waveforms.i * +waveforms.i)
                axes[axis_idx].plot(time, waveform_abs, linestyle, label="abs", color="green")

            bus = self.buses[axis_idx]
            if isinstance(bus.system_control, ReadoutSystemControl):
                plt.axvline(x=bus.acquire_time(idx=idx), color="red", label="Acquire time")
            axes[axis_idx].minorticks_on()
            axes[axis_idx].grid(which="both")

        plt.xlabel("Time (ns)")
        plt.legend(loc="upper right")
        plt.tight_layout()
        return figure

    def _waveforms_dict(self, modulation: bool = True, resolution: float = 1.0, idx: int = 0) -> dict[str, Waveforms]:
        """Get pulses of each bus.

        Args:
            resolution (float): The resolution of the pulses in ns.

        Returns:
            dict[str, Waveforms]: Dictionary containing a list of the I/Q amplitudes of the pulses applied on each bus.
        """
        return {bus.alias: bus.waveforms(modulation=modulation, resolution=resolution, idx=idx) for bus in self.buses}
